(* codegen.sml *)

signature CODEGEN =
sig
  structure F : FRAME
  structure R : REGISTER

  (* translate each canonical tree into a list of assembly instructions *)
  val codegen : Tree.stm -> Assem.instr list 

  (* converting a string fragment into the actual assembly code *)
  val string : Temp.label * string -> string

  (* procEntryExit sequence + function calling sequence tune-up 
   * + mapping pseudo-registers to memory load/store instructions 
   * and actual registers.
   * This is a post-pass, to be done after register allocation.
   *)
  val procEntryExit : {name : Temp.label, 
                          body : (Assem.instr * Temp.temp list) list,
                          allocation : R.register Temp.Table.table,
                          formals : Temp.temp list,
                          frame : Frame.frame} -> Assem.instr list

end (* signature CODEGEN *)

structure Codegen : CODEGEN =
struct

 structure T = Tree
 structure A = Assem
 structure Er = ErrorMsg
 structure F = Frame
 structure R = Register

 val instrs : A.instr list ref = ref []
 val push_live_instr = "# push live regs\n"
 val pop_live_instr  = "# pop live regs\n"

 fun bug s = Er.impossible ("bug in codegen: " ^ s)

 (* Prints a string fragment. All control characters are escaped
    and printed using their hexademical ASCII representation *)
 fun string (l, s) =
     let 
	 val label = Symbol.name l
	 val size  = Int.toString (String.size s)
	 fun escape c =
	     if ord c >= 32 then
		 str c
	     else
		 "\\x" ^ (Int.fmt StringCvt.HEX (ord c))
	 val string = String.translate escape s
     in
	 label ^ ":\t.long " ^ size ^ "\n" ^
	 "\t.ascii \"" ^ string ^ "\"\n"
     end

 (* add an instruction to the final instructions list *)
 fun emit instr = instrs := instr :: !instrs

 (* allocates a new temporary, and calls a function that emits code
    calculating the result into the temporary *)
 fun result f = 
     let
	 val r = Temp.newtemp ()
     in
	 (f r; r)
     end

 (* an auxiliary function to convert strings to integers. Needed
    because the default Int.toString SML function prints the negation
    sign as a tilde. *)
 fun intToString i = if i >= 0 then
			 Int.toString i
		     else
			 "-" ^ (Int.toString (~i))
			
 (* perform maximal munching on an expression *) 
 fun munchExp exp =
     case exp of

	 (* since addl/subl operate in place, allocate new temporary,
            store the one operand to it, and add to it or subtract from it
	    the second operand *)
	 (T.BINOP(binop as T.PLUS, e1, T.CONST(i2))
       | T.BINOP(binop as T.PLUS, T.CONST(i2), e1)
       | T.BINOP(binop as T.MINUS, e1, T.CONST(i2))) =>
	 let val t1 = munchExp e1
	     val opcode = case binop of T.PLUS => "addl" | T.MINUS => "subl"
	 in
	     result (fn r =>
			(emit(A.MOVE{assem="\tmovl `s0, `d0\n", src = t1, dst = r});
			 emit(A.OPER{assem="\t" ^ opcode ^ " $" ^ intToString i2 ^ ", `d0\n", src = [r], dst = [r], jump = NONE})))
	 end

       | (T.BINOP(binop as T.PLUS, e1, e2)
       | T.BINOP(binop as T.MINUS, e1, e2)) =>
	 let val t1 = munchExp e1
	     val t2 = munchExp e2
	     val opcode = case binop of T.PLUS => "addl" | T.MINUS => "subl"
	 in
	     result (fn r =>
			(emit(A.MOVE{assem="\tmovl `s0, `d0\n", src = t1, dst = r});
			 emit(A.OPER{assem="\t" ^ opcode ^ " `s0, `d0\n", src = [t2, r], dst = [r], jump = NONE})))
	 end

	 (* multiplication needs the first operand into EAX, and
            calculates the result into EDX:EAX, so both registers should
            be destinations of the imul instruction. *)
       | T.BINOP(T.MUL, e1, e2) =>
	 let
	     val t1 = munchExp e1
	     val t2 = munchExp e2
	 in
	     result (fn r =>
			(emit(A.MOVE{assem="\tmovl `s0, `d0\n", src=t1, dst=R.RV});
			 emit(A.OPER{assem="\timul `s0\n", src=[t2], dst=[R.RV,R.EDX], jump=NONE});
			 emit(A.MOVE{assem="\tmovl `s0, `d0\n", src=R.RV, dst=r})))
	 end
	 
         (* division needs the dividend into EDX:EAX, and calculates the 
            remainder into EAX and the quotient in EDX. Thus, we must be careful to
            clear EDX before doing the division. *)
       | T.BINOP(T.DIV, e1, e2) =>
	 let
	     val t1 = munchExp e1
	     val t2 = munchExp e2
	 in
	     result (fn r => (emit(A.OPER{assem="\tmovl $0, `d0\n", src=[], dst=[R.EDX], jump=NONE});
			      emit(A.MOVE{assem="\tmovl `s0, `d0\n", src=t1, dst=R.RV});
			      emit(A.OPER{assem="\tidiv `s0\n", src=[t2], dst=[R.RV,R.EDX], jump=NONE});
			      emit(A.MOVE{assem="\tmovl `s0, `d0\n", src=R.RV, dst=r})))
	 end

       | T.MEM(exp, 4) =>
	 result (fn r =>
		    emit(A.OPER{assem="\tmovl (`s0), `d0\n", src=[munchExp exp], dst=[r], jump=NONE}))

       | T.TEMP(t) =>
	 t

       | T.NAME(n) =>
	 result (fn r =>
		    emit(A.OPER{assem="\tmovl $" ^ Symbol.name n ^ ", `d0\n", src=[], dst=[r], jump=NONE}))

       | T.CONST(i) =>
	 result (fn r =>
		    emit(A.OPER{assem="\tmovl $" ^ intToString i ^ ", `d0\n", src=[], dst=[r], jump=NONE}))

	 (* function call: first, add a placeholder instruction to indicate where storing of the live caller-save
            registers should be done at procEntryExit time. Then, push all parameters in reverse order. Then,
            the function is called, and the returned result must be moved into the result register. The stack
            space holding the parameters should be reclaimed, and last, a placeholder for restoring the values
            of caller-save registers is added. *)
       | T.CALL(T.NAME(function), params) =>
	 let (* push a function parameter onto the stack *)
	     fun pushParam (T.NAME(lab)) =
		 emit(A.OPER{assem="\tpushl $" ^ Symbol.name lab ^ "\n", src=[], dst=[], jump=NONE})
	       | pushParam (T.CONST(i)) =
		 emit(A.OPER{assem="\tpushl $" ^ intToString i ^ "\n", src=[], dst=[], jump=NONE})
	       | pushParam e =
		 emit(A.OPER{assem="\tpushl `s0\n", src=[munchExp e], dst=[], jump=NONE})

             (* string containing the size of the parameters *)
	     val stackalloc = intToString (4 * (List.length params))

	 in
	     result (fn r =>
			(emit(A.OPER{assem=push_live_instr, src=[], dst=[], jump=NONE});
			 List.app pushParam (List.rev params);
			 emit(A.OPER{assem="\tcall " ^ Symbol.name function ^ "\n", src=[], dst=[R.RV], jump=NONE});
			 emit(A.MOVE{assem="\tmovl `s0, `d0\n", src=R.RV, dst=r});
			 emit(A.OPER{assem="\taddl $" ^ stackalloc ^ ", `d0\n", src=[R.SP], dst=[R.SP], jump=NONE});
			 emit(A.OPER{assem=pop_live_instr, src=[], dst=[], jump=NONE})))
	 end

       | _ => 
	 bug "expression not implemented"

 (* function to perform maximal munch on statements *)
 fun munchStm stm =
     let
	 fun relop rel = case rel of
			     T.EQ => "je"
			   | T.NE => "jne"
			   | T.LT => "jl"
			   | T.GT => "jg"
			   | T.LE => "jle"
			   | T.GE => "jge"
			   | _ => bug "relative operator not implemented"

	 fun f (T.SEQ(stm1,stm2)) =
	     (f stm1; f stm2)
	     
	   | f (T.LABEL(lab)) =
	     emit(A.LABEL{assem = Symbol.name lab ^ ":\n", lab = lab})
	     
	   | f (T.JUMP(T.NAME lab, labels)) =
	     (emit(A.OPER{assem = "\tjmp " ^ Symbol.name lab ^ "\n", src=nil, dst=nil, jump = SOME labels}))

	     (* conditional jump. x86 permits the first operand to be a constant or a register. *)
	   | f (T.CJUMP(T.TEST(rel, exp1, T.CONST(exp2)), labtrue, labfalse)) =
	     (emit(A.OPER{assem = "\tcmp $" ^ intToString exp2 ^ ", `s0\n", src=[munchExp exp1], dst=nil, jump = NONE});
	      emit(A.OPER{assem = "\t" ^ relop rel ^ " " ^ Symbol.name labtrue ^ "\n",
			  src=nil, dst=nil, jump = SOME [labtrue, labfalse]}))

	   | f (T.CJUMP(T.TEST(rel, exp1, exp2), labtrue, labfalse)) =
	     (emit(A.OPER{assem = "\tcmp `s1, `s0\n", src=[munchExp exp1, munchExp exp2], dst=nil, jump = NONE});
	      emit(A.OPER{assem = "\t" ^ relop rel ^ " " ^ Symbol.name labtrue ^ "\n",
			  src=nil, dst=nil, jump = SOME [labtrue, labfalse]}))


	     (* move instructions. x86 supports src: memory+offset, memory, register, immediate, and dst: memory+offset, memory, register.
                but we can't have both src and dst be memory addressing. *)

	   | f (T.MOVE(T.MEM(T.BINOP(T.PLUS,e1,T.CONST(offset)),4), T.CONST(i))) =
	     emit(A.OPER{assem = "\tmovl $" ^ (intToString i) ^ ", " ^ intToString offset ^ "(`s0)\n", src=[munchExp e1], dst=[], jump=NONE})

	   | f (T.MOVE(T.MEM(T.BINOP(T.PLUS,T.CONST(offset),e1),4), T.CONST(i))) =
	     emit(A.OPER{assem = "\tmovl $" ^ (intToString i) ^ ", " ^ intToString offset ^ "(`s0)\n", src=[munchExp e1], dst=[], jump=NONE})

	   | f (T.MOVE(T.MEM(T.BINOP(T.PLUS,e1,T.CONST(offset)),4), e2)) =
	     emit(A.OPER{assem = "\tmovl `s0, " ^ intToString offset ^ "(`s1)\n", src=[munchExp e2, munchExp e1], dst=[], jump=NONE})

	   | f (T.MOVE(T.MEM(T.BINOP(T.PLUS,T.CONST(offset),e1),4), e2)) =
	     emit(A.OPER{assem = "\tmovl `s0, " ^ intToString offset ^ "(`s1)\n", src=[munchExp e2, munchExp e1], dst=[], jump=NONE})

	   | f (T.MOVE(T.MEM(e1, 4), T.CONST(i))) =
	     emit(A.OPER{assem = "\tmovl $" ^ (intToString i) ^ ", (`s0)\n", src=[munchExp e1], dst=[], jump=NONE})

	   | f (T.MOVE(T.MEM(e1, 4), e2)) =
	     emit(A.OPER{assem = "\tmovl `s0, (`s1)\n", src=[munchExp e2,munchExp e1], dst=[], jump=NONE})

	   | f (T.MOVE(e1, T.MEM(T.BINOP(T.PLUS,e2,T.CONST(offset)),4))) =
	     emit(A.OPER{assem = "\tmovl " ^ intToString offset ^ "(`s0), `d0\n", src=[munchExp e2], dst=[munchExp e1], jump=NONE})

	   | f (T.MOVE(e1, T.MEM(T.BINOP(T.PLUS,T.CONST(offset),e2),4))) =
	     emit(A.OPER{assem = "\tmovl " ^ intToString offset ^ "(`s0), `d0\n", src=[munchExp e2], dst=[munchExp e1], jump=NONE})

	   | f (T.MOVE(e1, T.MEM(e2,4))) =
	     emit(A.OPER{assem = "\tmovl (`s0), `d0\n", src=[munchExp e2], dst=[munchExp e1], jump=NONE})

	   | f (T.MOVE(e1, T.CONST(i))) =
	     emit(A.OPER{assem = "\tmovl $" ^ intToString i ^ ", `d0\n", src=[], dst=[munchExp e1], jump=NONE})

	   | f (T.MOVE(e1, e2)) =
	     emit(A.MOVE{assem = "\tmovl `s0, `d0\n", src=munchExp e2, dst=munchExp e1})

	   | f (T.EXP(e1)) =
	     ignore(munchExp(e1))

	   | f _ = bug "cannot munch statement"
     in
	 f stm
     end

 (* given a canonical tree (which is just a statement), perform maximal
    munching on it, and return the resulting instruction list in reverse *)	 
 fun codegen tree =
     (instrs := [];
      munchStm tree;
      List.rev (!instrs))



(************************************************************
  The following is an example implementation of mapping pseudo-registers 
  to memory load/store instructions and actual registers.  It is done
  in a single pass.  It assumes that pseudo-register names start with
  the letter "f".  It uses the actual registers ECX and EDX as temporaries
  when a pseudo-register is an operand of an instruction.

  There is a special case that this function does NOT handle, but you MUST!
  The DIV instruction has special requirements.  Its dividend must be in EAX, 
  its divisor in a general-purpose register.  It returns both the quotient,
  in EAX, and the remainder, in EDX regardless where the original divisor was! 
  So be careful that a divide instruction does not trash something useful
  in EDX, and that you retrieve the correct resulut from the divide instruction.

  --- Changed to use ESI and EDI for pseudoregisters. That's because EDX interferes
      with division.
*)

  (* regname -- produce an assembly language name for the given machine
   * register or psuedo-register.
   * psuedo-registers are mapped to an expression for psuedo-register's
   * location in stack frame.
   *)
  (* regname : R.register -> string *)
  fun regname reg =
      if (String.isPrefix "f" reg) then
	  (* it's a psuedo-register *)
	  let
	      val (SOME prNum) = Int.fromString (String.extract(reg,1,NONE));
	      val offset = (prNum + 1) * 4
	  in
	      "-" ^ Int.toString(offset) ^ "(%ebp)"
	  end
      else
	  reg


 (* genSpills -- do our "poor man's spilling".  Maps all pseudo-register
  * references to actual registers, by inserting instructions to load/store
  * the pseudo-register to/from a real register
  *)
 fun genSpills (insns,saytemp) =
     let
	  (* doload() -- given name of a source register src, and a true
	   * machine register mreg, will return a load instruction (if needed)
	   * and a machine register.
	   *)
	  (* loadit: Temp.temp * Temp.temp -> string * Temp.temp *)
	  fun loadit (src,mreg) =
	      let 
		  val srcnm = (saytemp src)
	      in
		  if (String.isPrefix "f" srcnm) then
		      (* it's a fake register: *)
		      let
			  (* val _ = print ("loadit(): mapping pseudo-register `" ^ srcnm ^ "' to machine reg. `" ^ (saytemp mreg) ^"'\n"); *)
			  val loadInsn = "\tmovl\t" ^ (regname srcnm) ^ ", " ^ (saytemp mreg) ^ " # load pseudo-register\n"
		      in
			  (loadInsn, mreg)
		      end
		  else
		      (* no mapping needed *)
		      ("", src)
	      end
	  (* mapsrcs : produce a sequence of instructions to load
	   * pseudo-registers into real registers, and produce a list
	   * of sources which reflects the real registers.
	   *)
	  (* mapsrcs : Temp.temp list * Temp.temp list -> string * Temp.temp list *)
	  fun mapsrcs ([],_) = ("",[])
	    | mapsrcs (src::srcs,mreg::mregs) =
              let
                  val (loadInsn, src') = loadit(src,mreg)
                  val (loadRest, srcs') = mapsrcs(srcs,mregs)
              in
                  (loadInsn ^ loadRest, src'::srcs')
              end
	  (* findit -- like List.find, but returns SOME i, where i is index
	   * of element, if found
	   *)
          fun findit f l =
	      let
		  fun dosrch([],f,_) = NONE
		    | dosrch(el::els,f,idx) =
		      if f(el) then
			  SOME idx
		      else
			  dosrch(els,f,idx+1)
	      in
		  dosrch(l,f,0)
	      end

	  (* mapdsts -- after we have mapped sources to real machine
	   * registers, iterate through dsts.
	   * If dst is a pseudo-register then
	   *    if dst was also a src,
	   *         replace dst with machine register to which src is already
	   *         mapped
	   *    else
	   *         map dst to its own machine register (just use %ecx)
	   *    generate a store insn for dst to save result
	   *)
          (* mapdsts : Temp.temp list * Temp.temp list * Temp.temp list ->
	   *           string * Temp.temp list
	   *)
          (* N.B.!  This only works for dst of length 0 or 1 !! *)
          (* pre: length(dsts) <= 1 *)
	  fun mapdsts([],_,_) = ("",[])
	    | mapdsts(dst::dsts,srcs,newsrcs) =
	      let
		  val found = findit (fn e => e=dst) srcs
		  val dstnm = (saytemp dst)
	      in
		  if (isSome(found)) then
		      (* this dst is also a source *)
		      let
			  val idx=valOf(found)
			  val src=List.nth (srcs,idx)
			  val mreg=List.nth (newsrcs,idx)
		      in
			  if (src <> mreg) then
			      ("\tmovl\t`d0, " ^ (regname dstnm) ^ " # save pseudo-register\n", mreg::dsts)
			  else
			      (* no mapping *)
			      ("", dst::dsts)
		      end
		  else
		      (* this dst isn't a source, but it might be a pseudo-register *)
                      if (String.isPrefix "f" dstnm) then
                          (* it's a fake register: *)
                          (* we can safely just replace this destination with
                           * %ecx, and then write out %ecx to the pseudo-register
                           * location.  Even if %ecx was used to hold a different
                           * source pseudo-register, we won't end up clobbering
                           * it until after the source has been used...
                           *)
                          ("\tmovl\t`d0, " ^ (regname dstnm) ^ " # save pseudo-register\n", R.ESI::dsts)
                      else
                          (* no mapping *)
                          ("", dst::dsts)
	      end

	  fun mapInstr(A.OPER{assem=insn, dst=dsts, src=srcs, jump=jmp}) = 
	      let
		  val (loadinsns, newsrcs) = mapsrcs(srcs, [R.ESI, R.EDI])
                  val (storeinsns, newdsts) = mapdsts(dsts, srcs, newsrcs)
	      in
		  A.OPER{assem=loadinsns ^ insn ^ storeinsns,
			 dst=newdsts, src=newsrcs, jump=jmp}
	      end
	    | mapInstr(instr as A.LABEL _) = instr
	    | mapInstr(A.MOVE{assem=insn, dst=dst, src=src}) =
	      let
		  val (loadinsns, newsrcs) = mapsrcs([src], [R.ESI, R.EDI])
		  val (storeinsns, newdsts) = mapdsts([dst], [src], newsrcs)
	      in
		  A.OPER{assem=loadinsns ^ insn ^ storeinsns,
			 dst=newdsts, src=newsrcs, jump=NONE}
	      end
     in
         map mapInstr insns
     end


  (* procEntryExit sequence + function calling sequence tune-up 
   * + mapping pseudo-registers to memory load/store instructions 
   * and actual registers.
   * This is a post-pass, to be done after register allocation.
   *)
 fun procEntryExit {name, body, allocation, formals, frame={formals=_,offlst,locals,maxargs}} =
     let

	 (* how much space must be allocated on stack for local variables and pseudoregisters *)
	 val stackalloc = intToString (4 * (R.NPSEUDOREGS + !locals))

         (* function prologue: declare function symbol as global and add function label. On
            entering a function, the current frame pointer must be saved (control link), then 
            it should advance to the current top of the stack, and space must be allocated
            for locals by subtracting from the stack pointer. *)
	 val prologue =
	     [A.OPER {assem="\n.globl " ^ Symbol.name name ^ "\n", dst=[], src=[], jump=NONE},
	      A.LABEL{assem = Symbol.name name ^ ":\n", lab = name},
	      A.OPER {assem = "\tpushl `s0\n", src=[R.FP], dst=[], jump=NONE},
	      A.MOVE {assem = "\tmovl `s0, `d0\n", src=R.SP, dst=R.FP},
	      A.OPER {assem = "\tsubl $" ^ stackalloc ^ ", `d0\n", src=[R.SP], dst=[R.SP], jump=NONE}]

         (* epilogue: move the frame pointer back to the stack pointer, deallocating locals.
            Then, restore the previous frame pointer by popping the control link into it, and return
	    to the calling function. *)
	 val epilogue = 
	     [A.MOVE{assem = "\tmovl `s0, `d0\n", src=R.FP, dst=R.SP},
	      A.OPER{assem = "\tpopl `d0\n", src=[], dst=[R.FP], jump=NONE},
	      A.OPER{assem = "\tret\n", src=[], dst=[], jump=NONE}]

         (* lookup a temporary in the register allocation *)
	 fun lookup t = Option.valOf (Temp.Table.look(allocation, t))

         (* generate spill instructions for the pseudoregisters *)
	 fun spill body = genSpills(body, lookup)

         (* remove redundant move instructions (same source and destination registers) *)
	 fun notRedundant (A.MOVE{assem, src = a, dst = b}, _) =
	     lookup a <> lookup b
	   | notRedundant (_, _) = true

	 (* handle storing + restoring of caller save registers. Finds the placeholder
            instructions for these operations, and finds the live-in registers at the point
            where the corresponding function is called. These registers are pushed and popped
            as needed. *)
	 fun callerSaves instrs =
	     let

		 (* holds registers we just pushed *)
		 val regs = ref nil

                 (* push/pop a list of registers *)
		 val pushMany = List.foldl (fn (reg, cur) => cur ^ "\tpushl\t" ^ reg ^ "\n") ""
		 val popMany  = List.foldl (fn (reg, cur) => cur ^ "\tpopl\t" ^ reg ^ "\n") ""

                 (* given a list of instructions, find the live-in temporaries
                    at the first call instruction. This is calculated by finding
                    the live-out temporaries of the previous instruction. (We need
                    live-in and not live-out because the return-value register is always
                    live-out at the function call instruction (since it's read in the next
                    instruction *)
		 fun findLiveInCall nil = bug "can't find call after push live!"
		   | findLiveInCall ((il1 as (_, liveout)) ::
				     (rest as
					   (A.OPER{assem,...},_) :: rest')) =
		     (if String.isPrefix "\tcall" assem then
			  liveout
		      else
			  findLiveInCall rest)
		   | findLiveInCall (hd :: rest) = findLiveInCall rest

		 (* find registers that are allocated to temporaries that are live-in
                    at the first call instruction of a list of instructions. *)
		 fun findLiveRegsInCall instrs =
		     let
			 val livetemps = findLiveInCall instrs
			 fun inList l x = List.exists (fn y => lookup y = x) l
		     in
			 List.filter (inList livetemps) R.callersaves
		     end

		 (* discover the placeholders for pushing live registers and popping
                    live registers. When a push-live-registers placeholder is found, search
                    the following instructions for the relevant function call (it'll be the
                    first to find), to find the live registers at that point. These get stored
                    into regs, and when the pop-live-registers placeholder is found, regs gets
                    read so the relevant registers are popped in reverse. *)
		 fun f nil = nil
		   | f (l as
			  ((hd as (A.OPER {assem, src=[], dst=[], jump=NONE}), _)
			   :: rest)) =
		     
		     if assem = push_live_instr then
			 (regs := findLiveRegsInCall l;
			  (A.OPER{assem = pushMany (!regs), src=[], dst=[], jump=NONE} :: f rest))

		     else if assem = pop_live_instr then
			 (A.OPER{assem = popMany (List.rev (!regs)), src=[], dst=[], jump=NONE} :: f rest)

		     else
			 (hd :: f rest)

		   | f ((hd, _) :: rest) = hd :: f rest

	     in
		 f instrs
	     end


	 (* given an instruction list, we remove redundant moves, store and restore caller-save registers,
            and generate spills for pseudoregisters. Then, the final instruction list is ready. *)
	 val process = spill o callerSaves o (List.filter notRedundant)

     in

	 prologue @ process body @ epilogue

     end


end (* structure Codegen *)
