type pos = int

type svalue = Tokens.svalue
type ('a,'b) token = ('a,'b) Tokens.token
type lexresult  = (svalue,pos) token

val lineNum = ErrorMsg.lineNum
val linePos = ErrorMsg.linePos
fun err(p1,p2) = ErrorMsg.error p1

(* auxiliary variables and functions for handling comments and strings *)

(* commentsNest: ref int
   keeps the current level of nesting inside comments *)
val commentsNest = ref 0

(* insideString : ref bool
   are we currently processing a string literal or not *)
val insideString = ref false

(* currentString : ref string
   keeps the full string up until the current point in the file
   when inside a string literal *)
val currentString = ref ""

(* stringStart : ref int
   keeps the yypos where the current string literal started  *)
val stringStart = ref 0

(* reset: unit -> unit
   Should be called when a new file is fed to the lexer.
   Used to reset the comments nesting level and the inside string
   literal variable. *)
fun reset () = (commentsNest := 0; insideString := false)

(* EOF and EOLN handling functions *)

(* eof : unit -> Token.token
   Called on end-of-file by the lexer. Checks whether we are
   in an unterminated comment or string and issues error messages
   accordingly *)
fun eof () =
    let
	val pos = hd(!linePos)
    in
	(if !commentsNest <> 0 then
	    ErrorMsg.error pos "Unterminated comment at end-of-file."
	else if !insideString then
	    ErrorMsg.error pos "Unterminated string literal at end-of-file."
	else
	    ());
	Tokens.EOF(pos,pos)
    end

(* newline : int -> unit
   Should be called on newline to update line number counters. *)

fun newline yypos = 
    (lineNum := !lineNum+1; linePos := yypos :: !linePos)

(* updateNewlines : string -> int -> unit
   Should be called after a string with many
   newlines is processed without calling newline. The
   newlines contained in the string are counted, and the
   counters lineNum and linePos as updated as if all the
   newlines' positions was in the last newline that occured. *)
fun updateNewlines str pos =
    let

	(* count_occurences : char -> string -> int
           counts the occurrences of the character in the string *)
	fun count_occurrences chr str =
	    List.length (List.filter (fn c => c = chr) (String.explode str))

	(* last_occurrence : char -> string -> int
           returns the position of the last occurrence of a character
           in a string. When the character does not occur in the
           string, returns 0. *)
	local
	    val n = ref 0
	in
	fun last_occurrence chr str = 
	    List.foldl
		(fn (c, cur_n) => (n := !n + 1; if c = chr then !n else cur_n))
		0
		(String.explode str)
	end

        (* do_n_times : (unit -> unit) -> int -> unit
           calls a function n times *)
	fun do_n_times what 0 = ()
	  | do_n_times what n = (what(); do_n_times what (n-1))

	val nls = count_occurrences #"\n" str
	val lastnlpos = pos + (last_occurrence #"\n" str) - 1
    in
	 do_n_times (fn () => newline lastnlpos) nls
    end

%% 

%s COMMENT;
%s STRING;
%header (functor TigerLexFun(structure Tokens: Tiger_TOKENS));

digit  = [0-9];
letter = [a-zA-Z];
id     = {letter}({letter}|{digit}|_)*;
spaces = [\ \t]+;
formatting = [\ \t\n\v\f\r];
control = [\064-\095];

%%

<INITIAL,COMMENT>\n	=> (newline yypos; continue());

<INITIAL>{spaces}       => (continue());

<INITIAL>","	=> (Tokens.COMMA(yypos,yypos+1));
<INITIAL>":"	=> (Tokens.COLON(yypos,yypos+1));
<INITIAL>";"	=> (Tokens.SEMICOLON(yypos,yypos+1));
<INITIAL>"("	=> (Tokens.LPAREN(yypos,yypos+1));
<INITIAL>")"	=> (Tokens.RPAREN(yypos,yypos+1));
<INITIAL>"["	=> (Tokens.LBRACK(yypos,yypos+1));
<INITIAL>"]"	=> (Tokens.RBRACK(yypos,yypos+1));
<INITIAL>"{"	=> (Tokens.LBRACE(yypos,yypos+1));
<INITIAL>"}"	=> (Tokens.RBRACE(yypos,yypos+1));
<INITIAL>"."	=> (Tokens.DOT(yypos,yypos+1));
<INITIAL>"+"	=> (Tokens.PLUS(yypos,yypos+1));
<INITIAL>"-"	=> (Tokens.MINUS(yypos,yypos+1));
<INITIAL>"*"	=> (Tokens.TIMES(yypos,yypos+1));
<INITIAL>"/"	=> (Tokens.DIVIDE(yypos,yypos+1));
<INITIAL>"="	=> (Tokens.EQ(yypos,yypos+1));
<INITIAL>"<>"	=> (Tokens.NEQ(yypos,yypos+2));
<INITIAL>"<"	=> (Tokens.LT(yypos,yypos+1));
<INITIAL>"<="	=> (Tokens.LE(yypos,yypos+2));
<INITIAL>">"	=> (Tokens.GT(yypos,yypos+1));
<INITIAL>">="	=> (Tokens.GE(yypos,yypos+2));
<INITIAL>"&"	=> (Tokens.AND(yypos,yypos+1));
<INITIAL>"|"	=> (Tokens.OR(yypos,yypos+1));
<INITIAL>":="	=> (Tokens.ASSIGN(yypos,yypos+2));

<INITIAL>while	=> (Tokens.WHILE(yypos,yypos+5));
<INITIAL>for	=> (Tokens.FOR(yypos,yypos+3));
<INITIAL>to	=> (Tokens.TO(yypos,yypos+2));
<INITIAL>break	=> (Tokens.BREAK(yypos,yypos+5));
<INITIAL>let	=> (Tokens.LET(yypos,yypos+3));
<INITIAL>in	=> (Tokens.IN(yypos,yypos+2));
<INITIAL>end	=> (Tokens.END(yypos,yypos+3));
<INITIAL>function => (Tokens.FUNCTION(yypos,yypos+8));
<INITIAL>var	=> (Tokens.VAR(yypos,yypos+3));
<INITIAL>type	=> (Tokens.TYPE(yypos,yypos+4));
<INITIAL>array	=> (Tokens.ARRAY(yypos,yypos+5));
<INITIAL>if	=> (Tokens.IF(yypos,yypos+2));
<INITIAL>then	=> (Tokens.THEN(yypos,yypos+4));
<INITIAL>else	=> (Tokens.ELSE(yypos,yypos+4));
<INITIAL>do	=> (Tokens.DO(yypos,yypos+2));
<INITIAL>of	=> (Tokens.OF(yypos,yypos+2));
<INITIAL>nil	=> (Tokens.NIL(yypos,yypos+3));

<INITIAL>{id}	=> (Tokens.ID(yytext, yypos, yypos + size yytext));

<INITIAL>{digit}+ => (Tokens.INT(valOf (Int.fromString yytext),
                        yypos, yypos + size yytext));

<INITIAL>"/*" => (YYBEGIN COMMENT; commentsNest := 1; continue());
<INITIAL>"*/" => (ErrorMsg.error yypos ("unmatched */"); continue());
<COMMENT>"/*" => (commentsNest := !commentsNest + 1;
		  continue());
<COMMENT>"*/" => (commentsNest := !commentsNest - 1;
		  if !commentsNest = 0 then
		    (YYBEGIN INITIAL; continue())
		  else
		    continue());
<COMMENT>.    => (continue());

<INITIAL>"\"" => (YYBEGIN STRING; insideString := true; currentString := ""; stringStart := yypos; continue());
<STRING>"\""  => (YYBEGIN INITIAL; insideString := false; Tokens.STRING(!currentString, !stringStart, yypos+1));
<STRING>[^\"\\\n]+ => (currentString := !currentString ^ yytext; continue());
<STRING>"\\n" => (currentString := !currentString ^ "\n"; continue());
<STRING>"\\t" => (currentString := !currentString ^ "\t"; continue());
<STRING>"\\\"" =>(currentString := !currentString ^ "\""; continue());
<STRING>"\\\\" =>(currentString := !currentString ^ "\\"; continue());
<STRING>"\\"{digit}{3} =>
                 (let
		      val digits = substring(yytext,1,3)
		      val code = valOf (Int.fromString digits)
		  in
		      (if code < 256 then currentString := !currentString ^ (str (chr code))
		       else ErrorMsg.error yypos ("illegal ascii escape");
		       continue())
		  end);
<STRING>"\\^"{control} =>
                (let
		     val control = String.sub(yytext,2)
		     val char = chr (ord control - 64)
		 in
		     currentString := !currentString ^ (str char);
		     continue()
		 end);
<STRING>"\\"{formatting}+"\\" => (updateNewlines yytext yypos; continue());
<STRING>"\\^". => (ErrorMsg.error yypos ("illegal control character " ^ (substring (yytext,1,2))); continue());
<STRING>"\\".  => (ErrorMsg.error yypos ("illegal escape sequence '" ^ yytext ^ "'"); continue());
<STRING>\n     => (ErrorMsg.error yypos ("line break inside string literal");
                   newline yypos; continue());

.       => (ErrorMsg.error yypos ("illegal character " ^ yytext); continue());

