signature SEMANT =
sig
  type ir_code
  val transprog : Absyn.exp -> Frame.frag list
end

functor SemantGen (Register: REGISTER_STD) : SEMANT =

struct

structure A = Absyn
structure Translate = TranslateGen(Register)
structure E = EnvGen(Translate)
structure S = Symbol
structure T = Types
val error = ErrorMsg.error
val impossible = ErrorMsg.impossible
type ir_code = Translate.gexp



val stublabel = Temp.newlabel
val stubaccess = Translate.allocInRegister ()
val dummyExp = Translate.dummyExp
val dummyBreak = Temp.newlabel ()

(*val debug_print = print*)
val debug_print = ignore

(*************************************************************************
 *                       UTILITY FUNCTIONS                               *
 *************************************************************************)

fun curry f x y = f (x,y)

fun uncurry f (x,y) = f x y

fun flip f x y = f y x

(* Compare two types for "semi-equality".  A record type with a NIL is 
 * valid, but two NIL types is an error.  Returns the non-null option 
 * type if semi-equal, and NONE otherwise.
 *)
fun typeEq t1 t2 pos =
    case (t1,t2) of
      (T.RECORD _, T.NIL) => SOME t1
    | (T.NIL, T.RECORD _) => SOME t2
    | (T.NIL, T.NIL) => 
      (error pos "Nil must be used in a context where its type can be\
                 \ determined"; 
       NONE)
    | _ => if t1 = t2 then SOME t1 else NONE

(* Check for type equality, but requires an additional predicate p to
 * be satisfied, that is, p(t1) = true.  Returns an INT, and reports
 * the supplied error message at the given position if any. *)
fun checkEqP t1 t2 p pos err =
    case typeEq t1 t2 pos of
      SOME t => if p t1 then T.INT else (error pos err; T.INT)
    | NONE => (error pos err; T.INT)

(* Generic lookup function for a symbol table. 
 * err: error message to report in case of error
 * dflt: default value to return if value does not exist in the table
 * env: symbol table to look up
 * id: symbol to look up
 * pos: position to report an error
 *)
fun lookup err dflt env id pos =
    case S.look (env, id) of
      NONE => (error pos (err ^ "'" ^ S.name id ^ "'"); dflt)
    | SOME e => e

(* Shortcut to look up a type in the typing environment.
 * Returns UNIT if type is not found.  Does not resolve NAME entries. *)
val lookupT' = lookup "unbound type variable " T.UNIT

(* Shortcut to look up a type of a variable in the environment. 
 * Returns a dummy VARentry of type UNIT if symbol is not found *)
val lookupV = lookup "unbound identifier " 
                     (E.VARentry {access=stubaccess, ty=T.UNIT})

(**************************************************************************
 *                   TRANSLATING TYPE EXPRESSIONS                         *
 *                                                                        *
 *              transty : (E.tenv * A.ty) -> (T.ty * A.pos)               *
 *************************************************************************)
(* Given a type environment and a type expression, returns the translated
 * type. *)
fun transty tenv (A.ArrayTy (id, pos)) = 
    (T.ARRAY ((lookupT' tenv id pos), ref ()), pos)
  | transty tenv (A.RecordTy tfields) =
    let val trans = List.map (fn {name,typ,pos} => (name, lookupT' tenv typ pos))
                             tfields
    in 
      (T.RECORD (trans, ref ()), 0)
    end
  | transty tenv (A.NameTy (id, pos)) = 
    ((lookupT' tenv id pos), pos)

(* Resolve the "NAME" type that could occasionally returned from the 
 * type environment. *)
fun actual_ty tenv (T.NAME (sym, ref (SOME ty))) = actual_ty tenv ty
  | actual_ty tenv (T.NAME (sym, ref NONE)) = impossible "bug3"
  | actual_ty tenv ty = ty

(* Shortcut to look up a type in the type environment.  
 * Unlike lookupT', this version resolves NAME entries by calling actual_ty. *)
fun lookupT tenv id pos =
    actual_ty tenv (lookupT' tenv id pos)

(* Make sure t1 and t2 are compatible, and regardless of any errors, return
 * t3.  Prints out an error at the given pos if t1 and t2 are not compatible.
 *)    
fun assumeT (t1:T.ty) (t2:T.ty) (t3:T.ty) pos err =
    case typeEq t1 t2 pos of
      SOME t => t3
    | NONE => (error pos err; t3)


val gensym = 
    let val count = ref 0 in
      (fn nam => 
          (count := !count + 1;
           S.symbol (nam ^ Int.toString (!count))))
    end

(* environment that keeps track of variables defined 
 * implicitly for loop counters.  *)
type cenv = bool S.table

(* environment of NAME aliases -- used for checking cycles in type aliases *)
type nenv = A.ty S.table

(* Performs some context checks before type-checking: 
 * 1. 'break' can only occur in while/for
 * 2. Variables introduced by for cannot be assigned to
 * 3. Circular types through type aliases 
 * 4. No two functions in a sequence of mutually recursive functions
 *    can have the same name
 * 5. No two types in a sequence of mutually recursive types can 
 *    have the same name
 * 
 * exp: A.exp to check; usually the whole program
 * cenv: should be an empty environment
 * loop: indicates whether exp is in a loop.  
 *       At the top level of a program, this should be false.
 *)
fun context_check exp cenv loop = let 
  (* operates on a var *)
  fun auxV (A.SimpleVar (sym,pos)) = () (* ok *)
    | auxV (A.FieldVar (var,sym,pos)) = auxV var
    | auxV (A.SubscriptVar (var,exp,pos)) = (auxV var; auxE exp)

  (* operates on an exp *)
  and auxE (A.AssignExp {var=A.SimpleVar(sym,pos'),exp,pos}) = 
      (case S.look (cenv, sym) of
         NONE => () (* unbound variable; report error later *)
       | SOME true => error pos' "loop counter can't be assigned to"
       | SOME false => () (* ok *);
       auxE exp)
    | auxE (A.AssignExp {var=A.FieldVar(var,sym,pos'),exp,pos}) =
      (auxV var; auxE exp)
    | auxE (A.AssignExp {var=A.SubscriptVar(var,exp',pos'),exp,pos}) =
      (auxV var; auxE exp'; auxE exp)
    | auxE (A.ForExp {var={name,escape},lo,hi,body,pos}) = 
      (auxE lo; auxE hi; 
       (* enter a loop variable *)
       context_check body (S.enter (cenv, name, true)) true)
    | auxE (A.VarExp var) = auxV var
    | auxE (A.AppExp {args, ...}) = List.app auxE args
    | auxE (A.OpExp {left, right, ...}) = (auxE left; auxE right)
    | auxE (A.RecordExp {fields, ...}) = List.app (auxE o #2) fields
    | auxE (A.SeqExp expposlist) =  List.app (auxE o #1) expposlist
    | auxE (A.IfExp {test, then', else'=SOME else', ...}) = 
      (auxE test; auxE then'; auxE else')
    | auxE (A.IfExp {test,then',else'=NONE,...}) = 
      (auxE test; auxE then')
    | auxE (A.WhileExp {test, body, ...}) = 
      (auxE test; context_check body cenv true)
    | auxE (A.BreakExp pos) = 
      if loop then () else 
      error pos "'break' must be in 'while' or 'for'"
    | auxE (A.ArrayExp {size, init, ...}) = (auxE size; auxE init)
    | auxE A.NilExp = () (* don't care *)
    | auxE (A.IntExp _) = () (* don't care *)
    | auxE (A.StringExp _) = () (* don't care *)
    | auxE (A.LetExp {decs,body,pos}) = 
      context_check body (List.foldl auxD cenv decs) loop

  (* operates on a declaration in Let *)
  and auxD (A.FunctionDec fundecs, cenv) =
      (List.foldl 
           (fn ({name, params, body, pos, ...},seen) =>
               let val cenv' =
                       List.foldl 
                           (fn ({var={name,...},...}, cenv) =>
                               (* enter a normal variable *)
                               S.enter (cenv, name, false))
                           cenv params
               in 
                 (if seen name then 
                    error pos "functions cannot have the same name \
                              \in the same sequence"
                  else
                    ();
                  context_check body cenv' false;
                  (fn n => n=name orelse seen n))
               end)
           (fn n => false) fundecs;
       cenv)
    | auxD (A.VarDec {var={name,...}, init, ...}, cenv) = 
      (auxE init; S.enter (cenv, name, false) (* enter a normal var *))
    | auxD (A.TypeDec types, cenv) = 
      let val (nenv, nlist) =
              List.foldl 
                  (fn ({name,ty,pos}, (nenv,nlist)) =>
                      (if List.exists (curry op= name) nlist then
                        error pos "types cannot have the same name \
                                  \in the same sequence"
                       else 
                         ();
                       case ty of
                         A.NameTy (sym,pos) => 
                         (S.enter (nenv, name, ty), name::nlist)
                       | _ => (nenv, nlist)))
                  (S.empty, []) types
      in 
        List.app (cycle nenv) nlist; cenv
      end
      
  (* Go through NameTy paths in the name env and check for cycles given
   * a sym *)
  and cycle nenv sym =
      let fun aux start seen =
              case S.look (nenv, start) of
                NONE => () (* not in env; no cycles possible. *)
              | SOME (A.NameTy (s2,pos)) => (* found a name *)
                if seen s2 then
                  error pos "cycles in type detected"
                else 
                  aux s2 (fn s => s=s2 orelse seen s)
              | SOME _ => impossible "bug in cycle detection"
      in
        aux sym (fn s => s=sym)
      end
in 
  auxE exp
end (* end of context_check *)


fun getFunInfo name env = 
    case S.look (env, name) of
      SOME (E.FUNentry {level,label,...}) => (level, label)
    | _ => impossible "bug in getFunInfo"

(**************************************************************************
 *                   TRANSLATING EXPRESSIONS                              *
 *                                                                        *
 **************************************************************************)
fun transexp (env, tenv) expr lev break = let 
  fun g (A.OpExp {left, oper, right, pos}) = 
      let 
        val (e1, t1) = g left
        val (e2, t2) = g right
        val t =
            if List.exists (curry op= oper) 
                           [A.PlusOp, A.MinusOp, A.TimesOp, A.DivideOp] then
              assumeT t1 t2 T.INT pos
                      "Types of both operands should be int"
            else 
              if List.exists (curry op= oper)
                             [A.LtOp, A.LeOp, A.GtOp, A.GeOp] then
                checkEqP t1 t2 (fn t => t=T.INT orelse t=T.STRING) pos
                         "Types of both operands should be int or string"
              else
                assumeT t1 t2 T.INT pos
                        "Types of both operands do not match"
      in (Translate.mkbinop oper e1 t1 e2 t2, t)
      end

    | g (A.IntExp i) = (Translate.constInt i, T.INT)
    | g (A.StringExp (s,pos)) = (Translate.mkstring s, T.STRING)
    | g (A.NilExp) = (Translate.nilexp, T.NIL)
    | g (A.IfExp {test, then', else'=SOME else'', pos}) =
      let 
        val (e1, t1) = g test
        val (e2, t2) = g then'
        val (e3, t3) = g else''
        val t = 
        (assumeT t1 T.INT T.INT pos "condition clause must be an int"; 
         case typeEq t2 t3 pos of
           SOME t => t
         | NONE => (error pos "'then' and 'else' branches must have\
                              \the same type";
                    t2)) (* arbitrarily pick t2 so we can continue *)
      in (Translate.mkIf e1 e2 e3, t)
      end

    | g (A.IfExp {test, then', else'=NONE, pos}) =
      let 
        val (e1, t1) = g test
        val (e2, t2) = g then'
        val t = 
        (assumeT t1 T.INT T.INT pos "condition clause must be an int"; 
         assumeT t2 T.UNIT T.UNIT pos "'then' clause must produce no value")
      in (Translate.mkIf2 e1 e2, t)
      end

    | g (A.VarExp v) = h v
    | g (A.AppExp {func,args,pos}) = 
      (* look up func's types in the environment *)
      (case lookup "undefined function " 
                   (E.FUNentry {level=lev,label=stublabel(),
                                formals=[],result=T.UNIT})
                   env func pos of
         E.VARentry _ => (error pos ("symbol " ^ S.name func ^ 
                                     " is defined as a variable but \
                                     \is used as a function");
                          (dummyExp, T.UNIT))
       | E.FUNentry {level, label, formals, result} =>
         (* check if the argument types agree *)
         if List.length formals = List.length args then
           let val argExs =
                   ListPair.map 
                       (fn (arg, formalT) => 
                           let val (e, t) = (g arg) in
                             assumeT t (actual_ty tenv formalT) T.UNIT pos
                                     "argument type mismatch"; e                               
                           end)
                       (args, formals)
           in
             (Translate.mkApp label level lev argExs, result)
           end
         else 
           (error pos "invalid number of arguments"; (dummyExp, T.UNIT)))

    | g (A.SeqExp expposlist) = 
      let fun aux ((exp,pos),(es,_)) =
              let val (e,t) = g exp in (e::es, t) end
          val (exs, t) = List.foldl aux ([], T.UNIT) expposlist
      in
        (Translate.mkSeq (List.rev exs), t)
      end

    | g (A.AssignExp {var,exp,pos}) = 
      let val (v, tv) = h var
          val (e, te) = g exp
          val t = assumeT tv te T.UNIT pos "incompatible types in assignment"
      in 
        (Translate.assign v e, t)
      end        

    | g (A.WhileExp {test,body,pos}) = 
      let val (t, tt) = g test
          val (whlExp,tb) = Translate.mkWhile t (fn done => transexp (env,tenv) body lev done)
          val _ = assumeT tt T.INT T.INT pos "'test' clausemust be an int"
          val twhile = assumeT tb T.UNIT T.UNIT pos 
                               "body of while must produce no value"
      in
        (whlExp, twhile)
      end

    (* for loop is just syntactic sugar for "while". *)
    | g (A.ForExp {var=(i as {name,escape}),lo,hi,body,pos}) = 
      let val limit = gensym "*limit*"
          val limitV = A.VarExp (A.SimpleVar (limit, pos))
          val iV = A.VarExp (A.SimpleVar (name, pos))
          val decs = 
              [A.VarDec {var = i,
                         typ = SOME (S.symbol "int", pos),
                         init = lo, pos = pos},
               A.VarDec {var = {name = limit, escape = ref true},
                       typ = SOME (S.symbol "int", pos),
                       init = hi, pos = pos}]
          val test = 
              A.IfExp 
                  {test = A.OpExp {left=iV, oper=A.LtOp, 
                                   right=limitV, pos=pos}, 
                   then' = A.AssignExp {var = A.SimpleVar (name,pos), 
                                        exp = A.OpExp {left=iV, oper=A.PlusOp,
                                                       right=A.IntExp 1, pos=pos},
                                        pos = pos}, 
                   else' = SOME (A.BreakExp pos), 
                   pos = pos}
          val body = A.WhileExp {test = A.OpExp {left=iV, oper=A.LeOp, right=limitV, pos=pos}, 
                                 body = A.SeqExp [(body,pos), (test,pos)], pos = pos}
          val trans = A.LetExp {decs = decs, body = body, pos = pos} 
          val exp = A.IfExp {test = A.OpExp {left = lo, oper = A.LeOp, 
                                             right = hi, pos = pos},
                             then' = trans, else' = NONE, pos = pos}
      in
        g exp
      end
      (*
      let 
        val (elo, tlo) = g lo
        val (ehi, thi) = g hi
        val env'= S.enter (env, name, E.VARentry {access=stubaccess, ty=T.INT})
        val (ebody, tbody) = transexp (env', tenv) body lev
        val t = 
        (assumeT tlo T.INT T.INT pos "lower bound must be an int";
         assumeT thi T.INT T.INT pos "upper bound must be an int";
         assumeT tbody T.UNIT T.UNIT pos 
                 "body of FOR loop must produce no value")
      in (dummyExp, t)
      end
       *)

    | g (A.BreakExp pos) = (Translate.mkBreak break, T.UNIT)
    | g (A.LetExp {decs,body,pos}) = 
      let val (env, tenv, assignments) = transdecs (env,tenv,decs) lev
          val (e, t) = 
              transexp (env, tenv) body lev break
      in (Translate.mkLet assignments e, t)
      end

    | g (A.ArrayExp {typ,size,init,pos}) = 
      let val t = lookupT tenv typ pos in
        (case t of
           T.ARRAY (ty,_) => 
           let val ty = actual_ty tenv ty
               val (sizeE, sizeT) = g size
               val (initE, initT) = g init
           in
             assumeT sizeT T.INT T.INT pos "array size must be int";
             assumeT initT ty ty pos "initial value does not agree \
                                     \with declared type";
             (Translate.mkArray sizeE initE, t)
           end
         | _ => (error pos "must be an array type"; (dummyExp, T.UNIT)))
      end
      
    | g (A.RecordExp {typ,fields,pos}) =
      let val t = lookupT tenv typ pos in
        (case t of
           T.RECORD (symtylist, _) =>
           if List.length symtylist = List.length fields then
             let val gexps =
                     (ListPair.map 
                          (fn ((sym1,ty),(sym2,e,pos)) =>
                              if sym1 = sym2 then
                                let val ty = actual_ty tenv ty 
                                    val (gexp, t) = g e in
                                  (assumeT t ty ty pos 
                                           "initial value does not agree \
                                           \with record type";
                                           gexp)
                                end
                              else (error pos "field name does not match \
                                              \that of record type"; dummyExp)
                                   )
                          (symtylist, fields))
                 val recordexp = Translate.mkRecord gexps
             in
               (recordexp, t)
             end
           else
             (error pos "wrong number of fields in record \
                        \initialization"; (dummyExp, T.UNIT))
         | _ => (error pos "must be a record type"; (dummyExp, T.UNIT)))
      end

  (* function dealing with "var", may be mutually recursive with g *)
  and h (A.SimpleVar (sym,pos)) = 
      (case lookupV env sym pos of 
         E.VARentry {access, ty} => ((*debug_print ("varentry " ^ S.name sym);
                                     debug_print "\n";
                                     debug_print "in level...";
                                     Translate.debug_level lev;
                                     debug_print "\n";*)
                                     (Translate.simpleVar access lev, 
                                     actual_ty tenv ty))
       | E.FUNentry _ => (error pos ("function '" ^ S.name sym ^ 
                                     "' is used as a first-class value");
                          (dummyExp, T.UNIT)))
      
    | h (A.FieldVar (v,id,pos)) =
      (* make sure var is a record type *) 
      let val (e,tv) = h v in
        case tv of
          T.RECORD (symtylist, _) => 
          (* make sure id is in record *)
          let fun find ((sym,ty),(res,n)) =
                  if sym = id then 
                    (SOME (Translate.mkRecField e (Translate.constInt n),
                           actual_ty tenv ty), n+1)
                  else
                    (res, n+1)
              val (result,n) = List.foldl find (NONE, 0) symtylist
          in
            case result of
              NONE => (error pos ("field '" ^ S.name id ^ 
                                  "' does not exist in record");
                       (dummyExp , T.UNIT))
            | SOME r => r
          end
        | _ => (error pos "expecting a record type on lhs"; 
                (dummyExp, T.UNIT))
      end
    | h (A.SubscriptVar (v,exp,pos)) =
      (* make sure var is an array *)
      let val (ev,tv) = h v in
        case tv of
          T.ARRAY (ty,_) =>
          let val (es, ts) = g exp in
            (* make sure exp is an int *)
            (Translate.mkSubscript ev es,
             assumeT ts T.INT (actual_ty tenv ty) pos
                     "array subscript must be an int")
          end
        | _ => (error pos "expecting an array type on lhs"; (dummyExp, T.UNIT))
      end

in g expr
end



(**************************************************************************
 *                   TRANSLATING DECLARATIONS                             *
 **************************************************************************)
and transdec (env, tenv, A.VarDec {var={name,...}, typ=NONE, init, pos}) lev = 
    (* case where defining a value without specifying its type *)
    let val (e, t) = transexp (env, tenv) init lev dummyBreak (* !! *)
        val acc = Translate.allocInFrame lev
        val var = Translate.simpleVar acc lev
        val assignment = Translate.assign var e
    in 
      if t = T.NIL then
        (error pos "must specify record type for 'nil'"; (env, tenv, []))
      else
        (S.enter (env, name, E.VARentry {access=acc, ty=t}), tenv, [assignment])
    end
  | transdec (env, tenv, A.VarDec {var={name,...},typ=SOME (s,p1),init,pos}) lev =
    let 
      val (initE, initT) = transexp (env, tenv) init lev dummyBreak (* !! *)
      val consT = lookupT tenv s p1
      val exprT =
          case (initT, consT) of
            (T.NIL, T.RECORD _) => consT
          | (T.NIL, _) => (error pos "initializing expressions of type \
                                     \NIL must be constrained by a record \
                                     \type";
                           T.UNIT)
          | (t, u) => assumeT t u u pos
                              "type of initializing expression \
                              \is not compatible with claimed type"
      val acc = Translate.allocInFrame lev 
      val var = Translate.simpleVar acc lev
      val assignment = Translate.assign var initE
    in (S.enter (env, name, E.VARentry {access=acc, ty=exprT}), tenv, [assignment])
    end
    
  | transdec (env, tenv, A.FunctionDec(declist)) lev = 
    let 
      fun transresult NONE = T.UNIT
        | transresult (SOME (sym,pos)) = lookupT tenv sym pos

      (* Update the current env with arg types and return types of the 
       * functions *)
      val (newE, acclsts) = 
          
          List.foldl (fn ({name, params, result,...}, (env,acclsts)) =>
                         let val formalTs = 
                                 List.map 
                                     (fn {var, typ, pos} =>
                                         lookupT tenv typ pos)
                                     params
                           val resultT = transresult result
                           val _ = debug_print ("semant: making newlev for "^S.name name^"\n")

                           val (newlev, acclst) = Translate.newLevel {parent = lev, formals=
                                                            List.tabulate 
                                                            (List.length params, 
                                                             (fn _ => true))}
                         in (S.enter(env, name, 
                                    E.FUNentry {level = newlev, label = Translate.mkFunLbl (S.name name),
                                                formals = formalTs, 
                                                result = resultT}),
                             acclst::acclsts)
                         end)
                     (env,[]) declist
          
                     
    in
      (* ... and use the new env to type-check the function bodies *)
      ListPair.app 
          (fn ({name, params, result, body, pos, ...},acclst) =>
              let 
                (* formal parameters *)
                val augE =
                    ListPair.foldl
                        (fn ({var={name,...}, typ, pos}, accs, env) =>
                            S.enter (env, name, 
                                     E.VARentry {access = accs, 
                                                 ty = lookupT tenv typ pos}))
                          newE (params,acclst)
                val _ = debug_print ("entering function..."^S.name name ^"\n")
                val (funLevel, funLabel) = getFunInfo name newE
                val (bodyE, bodyT) = transexp (augE, tenv) body funLevel dummyBreak (* !! *)
                (* insert a proc fragment of this function *)
                val _ = Translate.mkFunc bodyE funLabel funLevel
                val resultT = transresult result
              in ignore 
                     (assumeT bodyT resultT resultT pos 
                              "type of function body does not match the \
                              \declared result type")
              end) (declist,List.rev acclsts);
      (* Finally, return the new env, and the original type env *)
      (newE, tenv, [])
    end
  | transdec (env, tenv, A.TypeDec(declist)) lev = 
    let val stubE = 
            List.foldl (fn ({name,ty,pos},tenv) =>
                           S.enter (tenv, name, T.NAME (name, ref NONE)))
                       tenv declist

    in
      (* go through each declaration and fill in the blanks in the 
       * stub environment *)
      List.app (fn {name,ty,pos} => 
                   let val (t,p) = transty stubE ty 
                   in 
                     case S.look (stubE, name) of
                       SOME (T.NAME (sym, ty)) => ty := SOME t
                     | _ => impossible "bug1"
                   end)
               declist;
      (env,stubE, [])
    end

(*** transdecs : (E.env * E.tenv * A.dec list) -> level -> (E.env * E.tenv) ***)
and transdecs (env,tenv,decs) lev =
    List.foldl (fn (dec,(env,tenv,assignments)) => 
                   let val (env', tenv', assignments') = 
                           transdec (env,tenv,dec) lev in
                     (env', tenv', assignments @ assignments')
                   end)
               (env,tenv,[]) decs



(*** transprog : A.exp -> Frame.frag list ***)
fun transprog prog =
    let val _ = Translate.reset ()
        val _ = context_check prog S.empty false
        (* create the level for tigermain *)
        val (level, _) = Translate.newLevel {parent = Translate.outermost, formals=[]}
        val (body, t) = transexp (E.base_env, E.base_tenv) prog level dummyBreak
        val _ = Translate.mkFunc body (Temp.namedlabel "tigermain") level
    in      
       Translate.getResult()
    end

    
end  (* structure Semant *)


