(* translate.sml *)

signature TRANSLATE = 
sig 
  type level
  type access
  type frag

  (* The top-most level under which all the library functions are defined *)
  val outermost : level
  
  (* Creates a new level given the parent level and a list of booleans, 
   * indicating whether each argument escapes.  In this particular 
   * implementation, all arguments are simply assumed to escape no matter 
   * what is given.  
   * 
   * Returns the newly created level and a list of "access" structure for
   * the formals.
   *)
  val newLevel : {parent: level, formals: 'a list} -> (level * access list)

  (* Allocates space for a local variable in the given level. 
   * Returns the access information for that variable.
   *)
  val allocInFrame : level -> access

  (* Allocates the variable in register.  Not implemented in this 
   * particular implementation. *)
  val allocInRegister : unit -> access

  (* Returns the lates list of itree fragments. *)
  val getResult : unit -> frag list

  (* Type for an itree expression. *)
  type gexp

  (* References a variable given the access info and the current level. *)
  val simpleVar : access -> level -> gexp

  (* Generates an itree expression for a constant int. *)
  val constInt : int -> gexp

  (* Inserts an PROC itree fragment for a function given its body, name, and level *)
  val mkFunc : gexp -> Temp.label -> level -> unit

  (* Creates an itree exp for a string literal *)
  val mkstring : string -> gexp

  (* Creates a binary op *)
  val mkbinop : Absyn.oper -> gexp -> Types.ty -> gexp -> Types.ty -> gexp

  (* Creates and initializes a record *)
  val mkRecord : gexp list -> gexp

  (* Creates and initializes an array *)
  val mkArray : gexp -> gexp -> gexp

  (* Creates an assignment exp *)
  val assign : gexp -> gexp -> gexp

  (* Creates a let expr with the assignments in a let expression and its body. *)
  val mkLet : gexp list -> gexp -> gexp

  (* Just a placeholder *)
  val dummyExp : gexp                  

  (* Creates a label for a function given its name. *)
  val mkFunLbl : string -> Temp.label                 

  (* An expr representing Nil. *)
  val nilexp : gexp

  (* Creates a function application given its label, callee's level, caller's level,
   * and a list of arguments. *)
  val mkApp : Temp.label -> level -> level -> gexp list -> gexp

  (* Creates an if expr *)
  val mkIf : gexp -> gexp -> gexp -> gexp
  val mkIf2 : gexp -> gexp -> gexp

  (* Creates a sequence of gexps given a list of them. *)
  val mkSeq : gexp list -> gexp

  (* Cleans up the itree fragments accumulated so far. *)
  val reset : unit -> unit

  (* Creates code that accesses an array subscript. *)
  val mkSubscript : gexp -> gexp -> gexp

  (* Creates code that accesses a record field. *)
  val mkRecField : gexp -> gexp -> gexp

  (* Creates a while loop. *)
  val mkWhile : gexp -> (Temp.label -> gexp*Types.ty)  -> (gexp * Types.ty)

  (* Creates a break expr given a label to jump to. *)
  val mkBreak : Temp.label -> gexp

  (* For debugging only *)
  val debug_level : level -> unit

end (* signature TRANSLATE *)


functor TranslateGen(Register : REGISTER_STD) : TRANSLATE = 
struct

  structure F = Frame
  structure Tr = Tree
  structure T = Temp
  structure Er = ErrorMsg

  datatype level = LEVEL of {frame : F.frame,              
                             sl_offset : int,
                             parent : level} * unit ref
                 | TOP

  (**** Debugging functions ****)
  fun print_level ({frame, sl_offset, parent}) =
      (print "level {";
       F.print_frame frame;
       print (", sl_offset = " ^ Int.toString sl_offset);
       print "}")

  fun debug_level (LEVEL (lev,_)) = print_level lev
    | debug_level TOP = print "TOP"


  (* The level in which the variable is defined, 
   * and the memory offset to the FP *)
  type access = level * int
  type frag = F.frag

  val sz = 4 (* the natural word size of the machine *)
  val FP = Register.FP

  val fragmentlist = ref ([] : frag list)

  fun addFrag frag = (fragmentlist := frag :: !fragmentlist)

  fun getResult () = !fragmentlist

  datatype gexp = Ex of Tr.exp
                | Nx of Tr.stm
                | Cx of T.label * T.label -> Tr.stm
  
  val outermost = TOP

  val dummyExp = Ex (Tr.CONST 42)

  (* error message for array out-of-bound error *)
  val arrayOOB = T.newlabel ()

  (* error message when a null pointer is dereferenced *)
  val nilDeref = T.newlabel ()

  fun mkstring' str lab = 
      (addFrag (F.DATA {lab=lab, s=str});
       Ex (Tr.NAME lab))

  fun mkstring str = mkstring' str (T.newlabel ())

  fun reset () = 
      (fragmentlist := [];
       mkstring' "Array index out of bounds\n" arrayOOB;
       mkstring' "Nil pointer dereferenced!?\n" nilDeref;
       ())
       

  fun allocInRegister () = (TOP,0) (* not implemented *)

  fun allocInFrame level =
      case level of 
        LEVEL ({frame,sl_offset,parent},_) => (level, F.allocInFrame frame)
      | TOP => Er.impossible "Translate.allocInFrame error"


  fun seq [] = Tr.EXP (Tr.CONST 0)
    | seq [x] = x
    | seq (x::xs) = Tr.SEQ (x, seq xs)


  (* Conversion functions between Ex, Cx, and Nx. *)
  fun unEx (Ex e) = e
    | unEx (Cx genstm) =
      let val r = T.newtemp ()
          val t = T.newlabel ()
          val f = T.newlabel ()
      in Tr.ESEQ (seq [Tr.MOVE (Tr.TEMP r, Tr.CONST 1),
                       genstm (t,f),
                       Tr.LABEL f,
                       Tr.MOVE (Tr.TEMP r, Tr.CONST 0),
                       Tr.LABEL t],
                  Tr.TEMP r)
      end
    | unEx (Nx s) = Tr.ESEQ (s, Tr.CONST 0) (* !! *)

  fun unCx (Ex (Tr.CONST 0)) = (fn (t,f) => Tr.JUMP (Tr.NAME f, [f]))
    | unCx (Ex (Tr.CONST 1)) = (fn (t,f) => Tr.JUMP (Tr.NAME t, [t]))
    | unCx (Ex e) = (fn (t,f) => Tr.CJUMP (Tr.TEST (Tr.EQ, Tr.CONST 0, e),
                                           f, t))
    | unCx (Cx c) = c
    | unCx (Nx _) = Er.impossible "cannot unCx an Nx"                              

  fun unNx (Nx s) = s
    | unNx (Ex e) = Tr.EXP e
    | unNx (cx as (Cx c)) = Tr.EXP (unEx cx)


  (* deals with a sequence of expressions. *)
  fun mkSeq exs = 
      let fun aux [] _ = Nx (Tr.EXP (Tr.CONST 0))
            | aux [exp] acc = Ex (Tr.ESEQ (seq (List.rev acc), unEx exp))
            | aux (x::xs) s = aux xs ((unNx x)::s)
      in
        aux exs []
      end


  fun newLevel {parent, formals} =
      (* extra one is for static link *)
      let val (fr, offsets) = F.newFrame (List.length formals + 1)  
          val lev = LEVEL ({frame = fr, sl_offset = List.hd offsets,
                            parent = parent}, ref ())
          val acclst = List.map (fn offset => (lev, offset)) (List.tl offsets)
      in (lev,acclst)
      end

  fun simpleVar (LEVEL (varLev,_), offset) thisLev = 
      let fun aux (LEVEL (lev,_)) acc =
              if lev = varLev then
                Tr.MEM (Tr.BINOP (Tr.PLUS, Tr.CONST offset, acc), sz)
              else (* follow the static link *)
                aux (#parent lev) 
                    (Tr.MEM (Tr.BINOP (Tr.PLUS, Tr.CONST (#sl_offset lev), acc), sz))
            | aux TOP acc = Er.impossible "TOP in aux in simpleVar"
      in Ex (aux thisLev (Tr.TEMP FP))
      end                        
    | simpleVar (TOP,_) TOP = Er.impossible "both levels are TOP"
    | simpleVar (TOP,_) _ = Er.impossible "first level is TOP"


  fun mkApp label calleeLev (callerLev as LEVEL ({frame,...},_)) args = 
      let val maxargs = #maxargs frame
          val _ = maxargs := Int.max (!maxargs, List.length args)
          fun mkApp (LEVEL (calleeLev, _)) =
              (* start with the current (caller's) level, 
               * trace back its parents until we find the callee's parent. *)
              let fun aux (lv as LEVEL (lev,_)) acc = 
                      if lv = #parent calleeLev then 
                        acc
                      else
                        aux (#parent lev) 
                            (Tr.MEM (Tr.BINOP (Tr.PLUS, Tr.CONST (#sl_offset lev), acc), sz))
                    | aux TOP acc = Er.impossible "TOP in aux"
                  val staticlink = aux callerLev (Tr.TEMP FP)        
              in
                Ex (Tr.CALL (Tr.NAME label, staticlink::(List.map unEx args)))
              end
            | mkApp TOP = 
              Ex (Tr.CALL (Tr.NAME label, List.map unEx args))
      in 
        mkApp calleeLev
      end
    | mkApp _ _ TOP _ = Er.impossible "mkApp: calling a function from TOP?"
                        
  (* makes a call to a library function *)
  fun externalCall s args =
      Tr.CALL (Tr.NAME (T.namedlabel s), args)

  (* When an unrecoverable error occurs, generate code that prints out
   * an error message and halts the program. *)
  val nilptrerr = 
      Nx (seq [Tr.EXP (externalCall "print" [Tr.NAME nilDeref]),
               Tr.EXP (externalCall "exit" [Tr.CONST 1])])

  val arrayerr = 
      Nx (seq [Tr.EXP (externalCall "print" [Tr.NAME arrayOOB]),
               Tr.EXP (externalCall "exit" [Tr.CONST 1])])


  val constInt = Ex o Tr.CONST

  val mkFunLbl = T.funlabel

  fun mkFunc body label (LEVEL ({frame, ...},_)) =
      let val frag = 
              F.PROC {name = label, 
                      body = Tr.MOVE (Tr.TEMP Register.RV, unEx body),
                      frame = frame}
      in 
        addFrag frag
      end
    | mkFunc body label TOP = Er.impossible "can't mkFunc under TOP"

  val nilexp = Ex (Tr.CONST 0)



  (* These functions translate a binop to its appropriate gexp. *)
  fun mkTest oper e1 e2 =
      let fun trrelop Absyn.EqOp = Tr.EQ
            | trrelop Absyn.NeqOp = Tr.NE
            | trrelop Absyn.LtOp = Tr.LT
            | trrelop Absyn.LeOp = Tr.LE
            | trrelop Absyn.GtOp = Tr.GT
            | trrelop Absyn.GeOp = Tr.GE
            | trrelop _ = Er.impossible "unsupported relop"
      in 
        Cx (fn (t,f) =>
               Tr.CJUMP (Tr.TEST (trrelop oper, e1, e2), t, f))
      end
                  
  and trbiop Absyn.PlusOp e1 e2 = Ex (Tr.BINOP (Tr.PLUS, e1, e2))
    | trbiop Absyn.MinusOp e1 e2 = Ex (Tr.BINOP (Tr.MINUS, e1, e2))
    | trbiop Absyn.TimesOp e1 e2 = Ex (Tr.BINOP (Tr.MUL, e1, e2))
    | trbiop Absyn.DivideOp e1 e2 = Ex (Tr.BINOP (Tr.DIV, e1, e2))
    | trbiop otherwise e1 e2 = mkTest otherwise e1 e2

  and complement e = Tr.BINOP (Tr.MINUS, Tr.CONST 1, e)

  and strEq e1 e2 = externalCall "stringEqual" [e1, e2]

  and strLt e1 e2 = externalCall "stringLessThan" [e1, e2]

  and strGt e1 e2 = strLt e2 e1

  and mkbinop oper e1 t1 e2 t2 = 
      let fun typ (t1 as (Types.RECORD _)) (Types.NIL) = t1
            | typ (Types.NIL) (t2 as (Types.RECORD _)) = t2
            | typ t1 t2 = if t1 = t2 then t1 
                          else t1 (*Er.impossible "type error in mkbinop"*) 
                               (* this is already handled in the type checker,
                                  so don't report the error again. *)

          fun mkbinop' e1 e2 Types.INT =
              trbiop oper e1 e2
            | mkbinop' e1 e2 Types.STRING = 
              (case oper of 
                 Absyn.EqOp => Ex (strEq e1 e2)
               | Absyn.NeqOp => Ex (complement (strEq e1 e2))
                                
               | Absyn.LtOp => Ex (strLt e1 e2)
               | Absyn.GeOp => Ex (complement (strLt e1 e2))
                               
               | Absyn.GtOp => Ex (strGt e1 e2)
               | Absyn.LeOp => Ex (complement (strGt e1 e2))
                               
               | _ => dummyExp (*Er.impossible "unsupported operator on strings"*))
            | mkbinop' e1 e2 (Types.RECORD _) = 
              if oper = Absyn.EqOp then trbiop oper e1 e2
              else dummyExp (*Er.impossible "invalid operator on records"*)
            | mkbinop' _ _ _ = dummyExp (*Er.impossible "unsupported type combination in mkbinop"*)
      in mkbinop' (unEx e1) (unEx e2) (typ t1 t2)
      end

  (* initialize a field in a record given:
   * e:      the expression to fill in
   * base:   the base address of the record
   * offset: the field offset 
   *)
  fun mkField e base offset =
      Tr.MOVE (Tr.MEM (Tr.BINOP (Tr.PLUS, Tr.TEMP base, Tr.CONST (offset*sz)), sz), e)

  fun mkRecord gexps =
      let val r = T.newtemp () (* the "base" address for the record *)
          (* store the value of each field in the record to the allocated space *)
          val (nFields,fields) =
              List.foldl (fn (ge,(offset,exs)) =>
                             (offset+1, (mkField (unEx ge) r offset) :: exs)) (0,[]) gexps
          (* allocate space for the record *)
          val alloc = 
              Tr.MOVE (Tr.TEMP r, externalCall "allocRecord" [Tr.CONST (nFields*sz)])
      in 
        Ex (Tr.ESEQ (seq (alloc::fields), Tr.TEMP r))
      end

  fun mkIf e1 e2 e3 =
      let val genstm = unCx e1
          val e2 = unEx e2
          val e3 = unEx e3
          val t = T.newlabel ()
          val f = T.newlabel ()
          val join = T.newlabel ()
          val r = T.newtemp () 
      in
        Ex (Tr.ESEQ (seq [genstm (t,f),
                          Tr.LABEL t, (* true branch *)
                          Tr.MOVE (Tr.TEMP r, e2), 
                          Tr.JUMP (Tr.NAME join, [join]), 
                          Tr.LABEL f, (* false branch *)
                          Tr.MOVE (Tr.TEMP r, e3),
                          Tr.LABEL join],
                     Tr.TEMP r))
      end

  fun mkIf2 e1 e2 = 
      let val genstm = unCx e1
          val e2 = unEx e2
          val t = T.newlabel ()
          val f = T.newlabel ()
      in
        Nx (seq [genstm (t,f),
                 Tr.LABEL t,
                 Tr.EXP e2,
                 Tr.LABEL f])
      end


  fun mkArray size init =
      let val a = T.newtemp()
          val s' = T.newtemp () (* used to store the size+1 value of the array *)
          val arraySZ = Tr.MOVE (Tr.TEMP s', Tr.BINOP (Tr.PLUS, unEx size, Tr.CONST 1))
          val e = unEx init
          (* allocate an array of size "size+1" *)
          val alloc = Tr.MOVE (Tr.TEMP a, externalCall "initArray" [Tr.TEMP s', e])
          (* tag the array with its size *)
          val tag = Tr.MOVE (Tr.MEM (Tr.TEMP a, sz), Tr.TEMP s') 
      in 
        Ex (Tr.ESEQ (seq [arraySZ, alloc, tag], Tr.TEMP a))
      end
       


  fun mkRecField record i =
      let val base = Tr.TEMP (T.newtemp())
          val movbase = Tr.MOVE (base, unEx record)
          val i' = unEx i
          val access = Ex (Tr.BINOP (Tr.PLUS, base, Tr.BINOP (Tr.MUL, i', Tr.CONST sz)))
      in
        Ex (Tr.MEM (Tr.ESEQ (seq [movbase],
                             unEx (mkIf (Ex base) access nilptrerr)), sz))
      end


  fun mkSubscript array i =
      let val base = Tr.TEMP (T.newtemp ())
          val movbase = Tr.MOVE (base, unEx array)
          val i' = Tr.TEMP (T.newtemp ())
          (* What we really want is the i+1'th cell *)
          val movi = Tr.MOVE (i', Tr.BINOP (Tr.PLUS, unEx i, Tr.CONST 1))
          val sizeminus1 = Tr.BINOP (Tr.MINUS, Tr.MEM (base, sz), Tr.CONST 1)
          val access = Ex (Tr.BINOP (Tr.PLUS, base, 
                                      Tr.BINOP (Tr.MUL, i', Tr.CONST sz)))
          val z = T.newlabel ()
          val cx = Cx (fn (t,f) => 
                          Tr.SEQ (Tr.CJUMP (Tr.TEST (Tr.GT, i', sizeminus1), t, z),
                                  Tr.SEQ (Tr.LABEL z, 
                                          Tr.CJUMP (Tr.TEST (Tr.LT, i', Tr.CONST 1),
                                                    t, f))))
      in
        Ex (Tr.MEM (Tr.ESEQ (seq [movbase, movi],
                             unEx (mkIf (Ex base)                 (* if base is not nil *)
                                        (mkIf cx arrayerr access) (* then do the access *)
                                        nilptrerr)) , sz))        (* else blow up *)
      end

  fun assign dst gexp =
      let val d = unEx dst
          val e = unEx gexp in
        Nx (Tr.MOVE (d, e))
      end


  fun mkLet [] body = body
    | mkLet assignments body = Ex (Tr.ESEQ (seq (List.map unNx assignments), unEx body))
      


  fun mkWhile cond body = 
      let val test = T.newlabel ()
          val done = T.newlabel ()
          val b    = T.newlabel ()
          val cond = unCx cond
          val (body',typ) = (body done)
      in
        (Nx (seq [Tr.LABEL test,
                  cond (b, done),
                  Tr.LABEL b,
                  unNx body', 
                  Tr.JUMP (Tr.NAME test, [test]),
                  Tr.LABEL done]),
         typ)
      end

  fun mkBreak break =
      Nx (Tr.JUMP (Tr.NAME break, [break]))
                         
end (* functor TranslateGen *)





     
