(* liveness.sml *)

signature LIVENESS =
sig

  datatype igraph = 
      IGRAPH of {graph : Graph.graph,
                 tnode : Graph.node Temp.Table.table,
                 gtemp : Temp.temp Graph.Table.table,
                 moves : (Graph.node * Graph.node) list}

   (* given a control-flow graph with data-flow information, this function
      returns a liveness-graph, a mapping table from temporaries to graph-nodes,
      a table for the reverse mapping, a list of move-edges, and a function that
      lists the live temporaries in each instruction *)
   val interferenceGraph : 
         Flow.flowgraph -> igraph * (Flow.Graph.node -> Temp.temp list)
  
   val show : TextIO.outstream * igraph -> unit

end (* signature LIVENESS *)

structure Liveness : LIVENESS = 
struct

  structure G = Graph
  structure F = Flow
  structure T = Temp

  fun bug s = ErrorMsg.impossible ("bug in liveness: "^s^"\n")

  datatype igraph = 
      IGRAPH of {graph : Graph.graph,
                 tnode : Graph.node Temp.Table.table,
                 gtemp : Temp.temp Graph.Table.table,
                 moves : (Graph.node * Graph.node) list}

  fun show (out, IGRAPH{graph, gtemp, ...}) = 
      DotGraph.dot(out, graph, fn node => Temp.makestring (valOf (G.Table.look(gtemp,node))))

  (* To construct the interference graph, it is convenient to
     construct a liveness map at each node in the FlowGraph first.
     For each node in the flowgraph, i.e., for each assembly 
     instruction, we want to easily look up the set S of live 
     temporaries. 
   *)

  (* the liveset (set of live temporaries) has a redundant representation:
     it holds both a table containing unit for each temporary in the set and
     a set-member list. This might not be space-optimal, but set manipulation
     functions can be very fast. *)
  type liveSet = unit Temp.Table.table * Temp.temp list

  (* the liveness map is a table that maps each flowgraph-node into a liveset *)
  type livenessMap = liveSet Flow.Graph.Table.table

  (* functions to manipulate the livesets. Property maintained: each temporary
     exists at most once in the table and in the list. *)

  (* empty liveset *)
  val livesetEmpty = (T.Table.empty, nil)

  (* tests whether a temporary is a member of the set *)
  fun livesetExists (set as (table,list)) temp = case T.Table.look(table,temp) of
						     NONE => false
						   | SOME _ => true

  (* add a temporary to a liveset. Maintain property that a member has only
     one entry in the table and the list by first checking if the temporary is already
     a member *)
  fun livesetAdd (set as (table,list)) temp =
      if livesetExists set temp then
	  set
      else
	  (T.Table.enter(table,temp,()), temp :: list)

  (* create a liveset from a list of temporaries *)
  fun livesetCreate list =
      List.foldl (fn (temp,set) => livesetAdd set temp) livesetEmpty list

  (* calculate the liveset that is the union of two other livesets *)
  fun livesetUnion set1 (set2 as (_,set2list)) =
      List.foldl (fn (temp,set) => livesetAdd set temp) set1 set2list

  (* calculate the set difference A - B = A intersect (complement B) *)
  fun livesetDiff (set1 as (_,set1list)) set2 =
      List.foldl (fn (temp,set) => livesetAdd set temp) livesetEmpty
		 (List.filter (not o livesetExists set2) set1list)

  (* tests two sets for equality (= A subset B and B subset A) *)
  fun livesetEqual (set1 as (_,set1list)) (set2 as (_,set2list)) =
      List.all (livesetExists set2) set1list andalso
      List.length set1list = List.length set2list

  (* gets the entry of a flowgraph-node out of a table mapping nodes to
     values (called flowtable), assuming that such an entry exists *)
  fun flowtableGet table node =
      Option.valOf (G.Table.look(table,node))
      handle Option.Option => bug "flowtableGet"

  (* computes livein and liveout sets for each flowgraph-node. Uses the simultaneous
     set equations method. The flow-graph is traversed in forward-order so this is not
     currently optimized *)
  fun computeLiveness (flowgraph as F.FGRAPH{control, def, use, ismove}) : (livenessMap * livenessMap) =
      let 
	  val nodes = G.nodes control

          (* re-order the nodes so as to make computation of liveness info faster
             (check Appel pg. 215) *)             

          (* table that maps nodes to their order *)
	  val orderedNodesTable = ref G.Table.empty
			     
          (* the order that the 'next' node will get *)
	  val currentOrder = ref 0

          (* order the node if it isn't already, and order its predecessors by depth-first search *)
	  fun addOrderFor node =
	      case G.Table.look(!orderedNodesTable, node) of
		  NONE => (orderedNodesTable := G.Table.enter(!orderedNodesTable,node,!currentOrder);
			   currentOrder := !currentOrder + 1;
			   List.app addOrderFor (G.pred node))
		| SOME _ => ()
			    
	  val _ = List.app addOrderFor (rev nodes)  (* actually only getOrdering (List.last nodes) should do *)

          (* now let's get a list of nodes in the order we want them *)
	  val orderedNodes = ListMergeSort.sort
				 (fn (n1, n2) =>
				     case (G.Table.look(!orderedNodesTable,n1), G.Table.look(!orderedNodesTable,n2)) of
					 (SOME o1, SOME o2) => o1 > o2
				       | _ => bug "orderedNodes")
			     nodes

          (* an empty livemap maps each node to an empty liveset *)
	  val livemapEmpty = List.foldl (fn (n,t) => G.Table.enter(t,n,livesetEmpty)) G.Table.empty nodes

          (* update the mapping of a node in a flowtable. If the mapping
             is kept the same, then the table returned is the same. Otherwise
             a table with the new mapping is returned and the change reference
             variable is set to true to reflect that a change in the flowtable was
             made. The table returned still contains the old mapping, but that's not
             a problem since Table.look will return the new mapping. *)
	  fun flowtableUpdate change table node newValue =
	      if livesetEqual (flowtableGet table node) newValue then
		  table
	      else
		  (change := true;
		   G.Table.enter(table,node,newValue))

          (* calculates livein and liveout sets, given initial livein and liveout.
             Updates the sets according to the liveness set equations and calls itself
             if they have been changed or returns current result if they haven't. *)
	  fun g livein liveout =
	      let
		  val change = ref false
                  (* updates liveness sets for a node using the equations:
                     livein[n] = use[n] union (liveout[n] diff def[n])
		     liveout[n] = union(forall nodes s in succ[n]) (livein[s]) *)
		  fun updateLiveness (node, (livein, liveout)) =
		      let
			  val setIn = flowtableGet livein node
			  val setOut = flowtableGet liveout node
			  val setDef = livesetCreate (flowtableGet def node)
			  val setUse = livesetCreate (flowtableGet use node)
			  val successors = G.succ node

			  (* updating first the liveout first, and using the new value in livein,
                             makes a tremendous increase in efficiency *)
			  val setOut' = (List.foldl (fn(s,c) => livesetUnion (flowtableGet livein s) c)
						    livesetEmpty successors)
			  val liveout' = flowtableUpdate change liveout node setOut'

			  val setIn' = livesetUnion setUse (livesetDiff setOut' setDef)
			  val livein' = flowtableUpdate change livein node setIn'
		      in
			  (livein', liveout')
		      end
		  val (livein', liveout') = List.foldl updateLiveness (livein, liveout) orderedNodes

	      in
		  if !change then
		      g livein' liveout'
		  else
		      (livein', liveout')
	      end
	  val (livein, liveout) = g livemapEmpty livemapEmpty

(* DEBUG: Prints livein and liveout sets *)
(* 	  fun livesetString (set as (_, list)) = *)
(* 	      List.foldl (fn (t,cur) => T.makestring t ^ "," ^ cur) "" list *)
(* 	  val _ = List.app (fn node => *)
(* 			       print ("Node: " ^ (G.nodename node) ^ " live-in:" ^ (livesetString (flowtableGet livein node)) ^ *)
(* 				      " live-out:" ^ (livesetString (flowtableGet liveout node)) ^ "\n")) *)
(* 			   nodes *)

      in

	  (livein, liveout)

      end

  (* after constructing the livenessMap, it is quite easy to
     construct the interference graph, just scan each node in
     the Flow Graph, add interference edges properly ... 
   *)


  fun interferenceGraph (flow as F.FGRAPH{control,def,use,ismove}) =
      let (* get live-out information for each flow-graph node *)
	  val (_, liveout) = computeLiveness flow

          (* tempgraph holds the interference graph. This graph contains a node for
             each temporary, and has edges between temporaries whose live-out ranges
             overlap  *)
	  val tempgraph = G.newGraph ()

          (* keep two mapping tables: from temporaries to interference-graph nodes and
             the reverse mapping *)
	  val tempToNodeTable = ref T.Table.empty
	  val nodeToTempTable = ref G.Table.empty

	  (* create a new node for a temporary in the interference-graph and update the
             mapping tables, only if it doesn't already exist *)
	  fun addTemp t1 =
	      case T.Table.look(!tempToNodeTable,t1) of
		  NONE =>
		  let val newnode = G.newNode tempgraph in
		      (tempToNodeTable := T.Table.enter(!tempToNodeTable,t1,newnode);
		       nodeToTempTable := G.Table.enter(!nodeToTempTable,newnode,t1);
		       newnode)
		  end
		| SOME n => n

          (* check whether an edge between t1 and t2 exists in the interference graph. since
             this graph is undirected we must check whether t1->t2 exists or t2->t1 exists. also,
             self-loops don't signify anything, so we assume that an edge exists from any node to
             itself *)
	  fun noEdgeExists t1 t2 =
	      if t1 = t2 then
		  false
	      else
		  case (T.Table.look(!tempToNodeTable, t1), T.Table.look(!tempToNodeTable, t2)) of
		      (SOME n1, SOME n2) => if List.exists (fn n' => (flowtableGet (!nodeToTempTable) n') = t2) (G.adj n1) then
						false
					    else
						true
		    | _ => true

          (* add an edge connecting temporaries t1 and t2. The edge is only added if no
             edge between them already exists. If a node in the interference graph for the
             temporaries has not yet been produced, such a node is created. *)
	  fun addEdge t1 t2 =
	      if noEdgeExists t1 t2 then
		  let val n1 = addTemp t1
		      val n2 = addTemp t2
		  in
		      G.mk_edge {from=n1, to=n2}
		  end
	      else
		  ()

          (* given a flowgraph node and a list of current move edges between nodes in the
             interference graph, it creates nodes and edges in the interference graph as needed (introducing
             nodes for temporaries not yet encountered), and returns a (possibly modified)
             list of move edges. When an instruction is a move instruction, no interference edges
             are added between the source and destination. *)
	  fun processNode (node, moves) =
	      let val ismove  = flowtableGet ismove node
		  val def     = flowtableGet def node
		  val use     = flowtableGet use node
		  val liveout = flowtableGet liveout node
                  (* liveoutList is the list of temporaries that are live-out in the current
                     instruction, minus the source temporaries if this is a move instruction. *)
		  val (_, liveoutList) = (if ismove then
					      livesetDiff liveout (livesetCreate use)
					  else
					      liveout)
		  fun getNode n = case T.Table.look(!tempToNodeTable,n) of
				      NONE => bug "can't get node of temporary"
				    | SOME t => t
	      in
		  ((* create an interference-graph node for all temporaries defined and used
                      in this instruction *)
		   List.app (ignore o addTemp) (def @ use);

		   (* add interference edges between temporaries defined at an instruction and
                      temporaries that are still live after the instruction *)
		   List.app (fn a => List.app (addEdge a) liveoutList) def;

		   (* if this is a move instruction, then add a move edge *)
		   if ismove then
		       case (def, use) of
			   ([to], [from]) => (getNode from, getNode to) :: moves
			 | _ => bug "move instruction with src and dest not single element lists"
		   else
		       moves)
		   
	      end

          (* finalize interference graph by producing nodes and edges, and get the list
             of move edges, by applying the previous function sequentially to all flowgraph nodes *)
	  val moves = List.foldl processNode nil (G.nodes control)

          (* a function that given a flowgraph node returns a list of temporaries
             that are live at that node. Used to determine which variables are still live
             at each instruction, so as not to save non-live variables in caller-save registers
             before function invocation. *)
	  fun liveout_converted node = 
	      case G.Table.look(liveout,node) of
		  NONE => bug "converted liveout"
		| SOME (t,l) => l

	  val igraph = IGRAPH{graph = tempgraph, tnode = !tempToNodeTable, gtemp = !nodeToTempTable,
			      moves = moves}

(* DEBUG: print interference graph *)
(* 	  val _ = show(TextIO.stdOut, igraph) *)

      in
	  (igraph, liveout_converted)
      end
		  

end (* structure Liveness *)

     

                 