(* makegraph.sml *)

signature MAKEGRAPH = 
sig

val instrs2graph : Assem.instr list -> Flow.flowgraph * Flow.Graph.node list

end

structure MakeGraph : MAKEGRAPH =
struct

structure A = Assem
structure F = Flow
structure G = F.Graph
structure T = G.Table

fun bug s = ErrorMsg.impossible ("bug in makegraph: "^s^"\n")
	      
(* The "instrs2graph" function takes a list of assembly instructions,
   and constructs its flowgraph and also returns the list of nodes in 
   the flowgraph. The instructions exactly correspond to the nodes in 
   the graph. If instruction m can be followed by instruction n (either
   by a jump or by falling through), there should be an edge from m to n
   in the graph.

   The flowgraph also maintains several attributes for each node in the 
   graph, i.e., the "def" set, the "use" set, and the "ismove" flag

 *)

fun instrs2graph instrs =
    let (* control holds the control-graph *)
	val control = G.newGraph()
        (* create a node for each instruction, and produce a list that
           keeps a pair of an instruction and its corresponding node *)
	val instrs' = List.map (fn i => (i, G.newNode control)) instrs
	val nodes = List.map #2 instrs' (* we could use G.nodes too, but this
                                           way we're sure instruction list and nodes list
					   have the same order *)

	(* searches for an instruction with a certain label and returns its
           corresponding node *)
	fun findLabelNode searchLabel =
	    let fun isSearchLabel (A.LABEL{assem, lab}, _) = searchLabel = lab
		  | isSearchLabel _ = false
		val find = List.find isSearchLabel instrs'
	    in
		case find of
		    SOME (instr,node) => node
		  | NONE => bug "findLabelNode -- label couldn't be found"
	    end

        (* for each instruction that contains a list of possible branching labels,
           add edges between the corresponding nodes in the graph. Assumes that
           the fall-through instruction is included in the branching labels list.
           Returns whether the instruction contains a list of branches or not. *)
	fun addJumpEdges (instr as (A.OPER{jump = SOME labels, ...}, node)) =
	    (List.app (fn targetLabel =>
			  G.mk_edge {from = node,
				    to = findLabelNode targetLabel})
		      labels; true)
	  | addJumpEdges _ = false

        (* takes an instruction list and adds edges from one statement to the next,
           except when an instruction explicitly defines branching labels (in which
           case appropriate branching edges are added) *)
	fun addNextEdges ((fst,fstnode) :: (snd,sndnode) :: tl) =
	    (if addJumpEdges (fst,fstnode) then
		 ()
	     else
		 G.mk_edge {from = fstnode, to = sndnode};
	     addNextEdges ((snd,sndnode)::tl))
	  | addNextEdges ((fst,fstnode) :: nil) = ignore(addJumpEdges (fst,fstnode))
	  | addNextEdges nil = ()

        (* finalize control-flow graph by adding edges for all instructions *)
	val _ = addNextEdges instrs'

        (* given an instruction-node list and the current sets of data-flow sets
           (definition, use and moves), it produces new versions of the sets, augmented
           accordingly *)
	fun augmentDefUseMove ((A.OPER{dst,src,...},node),(def,use,move)) =
	    (T.enter(def,node,dst),
	     T.enter(use,node,src),
	     T.enter(move,node,false))
	  | augmentDefUseMove ((A.LABEL{...},node),(def,use,move)) =
	    (T.enter(def,node,nil),
	     T.enter(use,node,nil),
	     T.enter(move,node,false))
	  | augmentDefUseMove ((A.MOVE{dst,src,...},node),(def,use,move)) =
	    (T.enter(def,node,[dst]),
	     T.enter(use,node,[src]),
	     T.enter(move,node,true))

        (* produce the final def, use, and ismove data-flow sets by applying the previous
           function for all instructions, starting with empty sets *)
	val (def, use, ismove) = List.foldl augmentDefUseMove
				 (T.empty, T.empty, T.empty) instrs'

(* DEBUG: print control-flow graph with data-flow information *)
(* 	val nodesToInstrTable = List.foldl (fn ((i,n),t) => T.enter(t,n,i)) T.empty instrs' *)
(* 	fun nodeToInstr n = *)
(* 	    (case T.look(nodesToInstrTable,n) of *)
(* 		 SOME i => i *)
(* 	       | NONE => bug "could not map node to instruction") *)
(* 	fun stringOfNode node = *)
(* 	    let val instr = nodeToInstr node *)
(* 		val assem = A.format Temp.makestring instr *)
(* 		val def = Option.valOf (T.look(def,node)) *)
(* 		val use = Option.valOf (T.look(use,node)) *)
(* 		val move = Option.valOf (T.look(ismove,node)) *)
(* 		val templist = fn (t,"") => (Temp.makestring t) *)
(* 				| (t,s)   => (Temp.makestring t ^ "," ^ s) *)
(* 		val defstring = List.foldl templist "" def *)
(* 		val usestring = List.foldl templist "" use *)
(* 		val movestring = if move then "true" else "false" *)
(* 	    in *)
(* 		"node: " ^ (G.nodename node) ^ " instr:" ^ assem ^ " [def:"^defstring^"] [use:"^usestring^"] [ismove:"^movestring^"]" *)
(* 	    end *)
(* 	val _ = DotGraph.dot(TextIO.stdOut,control,stringOfNode) *)

    in
	(F.FGRAPH{control = control,
		  def = def,
		  use = use,
		  ismove = ismove},
	 nodes)
    end
	
end