(* regalloc.sml *)

signature REG_ALLOC =
sig
   structure R : REGISTER_STD
   
   type allocation = Register.register Temp.Table.table

   val color : {interference : Liveness.igraph,
                initial : allocation,
                registers : R.register list} -> allocation

end (* signature REG_ALLOC *)

functor RegAllocGen(Register : REGISTER_STD) : REG_ALLOC =
struct
   structure R : REGISTER_STD = Register
   structure G = Graph
   structure I = Liveness
   structure T = Temp

   type allocation = R.register Temp.Table.table

   fun bug s = ErrorMsg.impossible ("bug in regalloc: "^s)

   (* The color function talkes an initial allocation table (which assigns
      temporary variables such as FP or SP into certain fixed registers)
      plus an interference graph and a list of registers, and returns
      a new allocation table (mapping from temporaries to registers).

      Only the simplify and select stages are implemented -- no spilling
      or coalescing is done.
    *)

   fun color {interference, initial, registers} =
       let
	   val I.IGRAPH {graph, gtemp, tnode, moves} = interference
	   val nodes = G.nodes graph

           (* a function that returns a mapping of a temporary in
              a temporary mapping table, assuming such a mapping exists *)
	   fun tempTlookup table temp = 
	       case T.Table.look(table,temp) of
		   NONE => bug "tempTlookup"
		 | SOME a => a

           (* the same thing for node mapping tables *)
	   fun nodeTlookup table node = 
	       case G.Table.look(table,node) of
		   NONE => bug "nodeTlookup"
		 | SOME a => a
           
           (* return the temporary that corresponds to an interference-graph node *)
	   val nodeToTemp = nodeTlookup gtemp

           (* a "removed" structure holds whether an interference-graph node is
              removed from the graph according to the current simplification *)
	   type removed = unit G.Table.table
 
           (* functions to handle the removed data structure *)
           
           (* empty removed set -- before any simplification *)
	   val removedEmpty = G.Table.empty
 
           (* remove an interference node *)
	   fun removedAdd rem what = G.Table.enter(rem,what,())

           (* check whether an interference node is currently removed or not *)
	   fun removedExists rem what = Option.isSome (G.Table.look(rem,what))

           (* check whether a node is already colored in the current allocation *)
	   fun allocExists alloc node = Option.isSome (T.Table.look(alloc,nodeToTemp node))

           (* the simplify step of interference-graph coloring: removes the first node
              found that isn't already removed, isn't already colored, and has degree less
              than the number of colors (registers) available. Such a node can always be
              colored trivially if the subgraph produced can be colored. *)
	   fun simplify removed alloc =
	       let (* calculate the degree of a node, not taking into account the removed nodes *)
		   fun degree node = 
		       List.length (List.filter (not o removedExists removed) (G.adj node))
 
 	           (* function that returns true if the properties described above hold for
                      a node *)
		   fun filterFunc k node = 
		       not (removedExists removed node) andalso
		       not (allocExists alloc node) andalso
		       degree node < k

                   (* a list of nodes that can be simplified *)
		   val possible = List.filter (filterFunc (List.length registers)) nodes
	       in
		   case possible of
		       nil => NONE
		     | hd :: tl => SOME hd
	       end

           (* this function returns the first available color for a node that
              can be trivially colored. It gets the colors of the currently colored neighbors
              of the node (not taking into account the currently removed notes), and returns
              one color that's not part of that set. *)
	   fun colorOne node removed alloc =
	       let (* current neighbors *)
		   val adj = List.filter (not o removedExists removed) (G.adj node)
			     
                   (* keep just the colored neighbors *)
		   val adj' = List.filter (allocExists alloc) adj

                   (* get the colored neighbors' colors *)
		   val adjRegs = List.map (tempTlookup alloc o nodeToTemp) adj'

                   (* function that checks whether something is a member of a list *)
		   fun existsInList l what = List.exists (fn t => t = what) l

                   (* the set of registers minus the set of registers allocated to the
                      current colored neighbors *)
		   val possible = List.filter (not o existsInList adjRegs) registers
	       in
		   case possible of
		       nil => bug "not enough registers -- should spill!"
		     | hd :: tl => hd
	       end

           (* try to trivially color all non-colored nodes not taking into account removed
              nodes *)
	   fun colorTrivial removed alloc =
	       let (* nodes that aren't removed and aren't already allocated *)
		   val notAlloced = List.filter (fn n => not (removedExists removed n)
							 andalso not (allocExists alloc n)) nodes

                   (* given a uncolored node and an allocation, try to color that node and
                      produce a new allocation containing that *)
		   fun allocOne (node,alloc) = T.Table.enter(alloc,nodeToTemp node,colorOne node removed alloc)
	       in
		   List.foldl allocOne alloc notAlloced
	       end

           (* the select step of the interference-graph coloring procedure: remove a node
              from the graph using the simplify step and call select recursively to color that. If no
              simplifications can be done, then try to color the graph trivially. That process should
              return a full allocation. *)
	   fun select removed alloc =
	       case (simplify removed alloc) of
		   NONE =>
		   let val alloc' = colorTrivial removed alloc
		       val num_alloc = List.length (List.filter (allocExists alloc') nodes)
		       val num_removed = List.length (List.filter (removedExists removed) nodes)
		       val num_nodes = List.length nodes
		   in
		       (if num_alloc < num_nodes - num_removed then
			    bug "could not simplify -- shouldn't happen"
			else
			    alloc')
		   end
		 | SOME newRemoved =>
		   let val alloc' = select (removedAdd removed newRemoved) alloc
		       val alloc'' = colorTrivial removed alloc'
		   in
		       alloc''
		   end

           (* produce a full allocation by applying select to the complete graph,
              starting with the initial allocation *)
	   val allocation = select removedEmpty initial

(* DEBUG: print register allocation *)
(* 	   fun printNode n = *)
(* 	       let val t = nodeToTemp n *)
(* 		   val reg = tempTlookup allocation t *)
(* 		   val t_string = Temp.makestring t *)
(* 		   val reg_string = reg *)
(* 	       in *)
(* 		   print ("Temporary " ^ t_string ^ " assigned to " ^ reg_string ^ "\n") *)
(* 	       end *)
(* 	   val _ = List.app printNode nodes *)
       in
	   allocation
       end

end (* functor RegAllocGen *)
