(* main.sml *)

signature MAIN = 
sig 
  val compile : string -> unit
  val mycompile : string -> unit
end (* signature MAIN *)

structure Main : MAIN = 
struct
  structure Semant = SemantGen(Register)
  structure RegAlloc = RegAllocGen(Register)

  structure C = Codegen
  structure F = C.F

  fun emitproc out (F.DATA {lab, s}) = TextIO.output(out,C.string(lab,s))

    | emitproc out (F.PROC{name, body, frame}) =
        let val stms = Canon.linearize body
            val stms' = Canon.traceSchedule(Canon.basicBlocks stms)
	    
	    val instrs = List.concat(map C.codegen stms') 

	    val (flowgraph, flownodes) = MakeGraph.instrs2graph instrs
	    val (interference, flownodeToLiveTemps) = Liveness.interferenceGraph flowgraph
	    fun pseudoregs 0 = []
	      | pseudoregs n = ("f" ^ (Int.toString(n-1))) :: pseudoregs (n-1)
	    val reglist = ["%eax", "%ebx", "%ecx", "%edx"] @ (pseudoregs Register.NPSEUDOREGS)

	    val initial = List.foldl (fn ((t,r),table) => Temp.Table.enter(table, t, r))
				     Temp.Table.empty
				     [ (Register.RV, "%eax"), (Register.EDX,"%edx"),
				       (Register.SP, "%esp"), (Register.FP, "%ebp"),
				       (Register.ESI,"%esi"), (Register.EDI,"%edi")]

	    val regalloc = RegAlloc.color {interference = interference, initial = initial,
					   registers = reglist}
			  
            val format0 = Assem.format (fn t => Option.valOf (Temp.Table.look(regalloc,t)))
	    val bodyWithLiveness = ListPair.zip (instrs,List.map flownodeToLiveTemps flownodes)
	    val instrs' = C.procEntryExit { name = name, body = bodyWithLiveness,
					    allocation = regalloc, formals = nil,
					    frame = frame }

         in app (fn i => TextIO.output(out,format0 i)) instrs'
        end


  fun debug_emitproc out (F.DATA {lab, s}) = TextIO.output(out,C.string(lab,s))

    | debug_emitproc out (F.PROC{name, body, frame}) =
        let val stms = Canon.linearize body
            val stms' = Canon.traceSchedule(Canon.basicBlocks stms)
	    
	    val instrs = List.concat(map C.codegen stms') 

	    val (flowgraph, flownodes) = MakeGraph.instrs2graph instrs
	    val (interference, flownodeToLiveTemps) = Liveness.interferenceGraph flowgraph
	    fun pseudoregs 0 = []
	      | pseudoregs n = ("<|" ^ (Int.toString(n-1)) ^ "|>") :: pseudoregs (n-1)
	    val reglist = ["%eax", "%ebx", "%ecx", "%edx"] @ (pseudoregs Register.NPSEUDOREGS)

	    val initial = List.foldl (fn ((t,r),table) => Temp.Table.enter(table, t, r))
				     Temp.Table.empty
				     [ (Register.RV, "%eax"), (Register.EDX,"%edx"),
				       (Register.SP, "%esp"), (Register.FP, "%ebp"),
				       (Register.ESI,"%esi"), (Register.EDI,"%edi")]

	    val regalloc = RegAlloc.color {interference = interference, initial = initial,
					   registers = reglist}
			  
            val format0 = Assem.format (fn t => Option.valOf (Temp.Table.look(regalloc,t)))
	    val bodyWithLiveness = ListPair.zip (instrs,List.map flownodeToLiveTemps flownodes)
	    val instrs' = C.procEntryExit { name = name, body = bodyWithLiveness,
					    allocation = regalloc, formals = nil,
					    frame = frame }

         in app (fn i => TextIO.output(out,format0 i)) instrs'
        end


  fun withOpenFile fname f = 
        let val out = TextIO.openOut fname
         in (f out before TextIO.closeOut out) 
               handle e => (TextIO.closeOut out; raise e)
        end 

  fun compile filename = 
        let val frags = Semant.transprog(Parse.parse filename)
         in withOpenFile (filename ^ ".s") 
                 (fn out => (app (emitproc out) frags))
        end

  fun mycompile filename = 
        let val frags = Semant.transprog(Parse.parse filename)
	    val out = TextIO.stdOut
         in (app (debug_emitproc out) frags)
        end

end (* structure Main *)


